import typing
from collections import OrderedDict

if typing.TYPE_CHECKING:
    from .device.block import BlockDevice

__all__ = [
    "BlockCache",
]

DEFAULT_CACHE_SIZE = 1024  # 1024 blocks


class BlockCache:
    dev: "BlockDevice"
    cache: OrderedDict[int, bytes]
    max_size: int

    def __init__(self, dev: "BlockDevice", max_size: int = DEFAULT_CACHE_SIZE):
        self.dev = dev
        self.cache = OrderedDict()
        self.max_size = max_size

    def read_block(self, block_number: int = 0) -> bytes:
        if block_number in self.cache:
            # Mark the block as most recently used
            self.cache.move_to_end(block_number)
            return self.cache[block_number]

        block_data = self.dev.read_block(block_number)
        self.cache[block_number] = block_data
        # Mark the block as most recently used
        self.cache.move_to_end(block_number)

        # If cache exceeds max_size, remove the least recently used item
        if len(self.cache) > self.max_size:
            self.cache.popitem(last=False)

        return block_data

    def write_block(self, buffer: bytes, block_number: int) -> None:
        self.dev.write_block(buffer, block_number)
        self.cache[block_number] = buffer
        self.cache.move_to_end(block_number)

        # If cache exceeds max_size, remove the least recently used item
        if len(self.cache) > self.max_size:
            self.cache.popitem(last=False)



from datetime import date

from xferx.pdp8.tss8fs import (
    TSS8_BLOCK_SIZE_BYTES,
    TSS8Filesystem,
    date_to_tss8,
    tss8_to_date,
)
from xferx.shell import Shell

DSK = "tests/dsk/tss8.dsk"


def test_tss8_to_date_basic():
    val = ((1980 - 1974) * 372) + ((5 - 1) * 31) + (15 - 1)
    assert tss8_to_date(val) == date(1980, 5, 15)


def test_date_round_trip():
    dates = [date(1974, 1, 1), date(1974, 1, 5), date(1974, 5, 10), date(1980, 6, 15), date(1982, 11, 30)]
    for d in dates:
        encoded = date_to_tss8(d)
        decoded = tss8_to_date(encoded)
        assert decoded == d


def test_tss8():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount t: /tss8 {DSK}.mo", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, TSS8Filesystem)

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir/brief t:[10,20]", batch=True)
    shell.onecmd("cd t:[11,21]", batch=True)
    shell.onecmd("dir t:", batch=True)
    shell.onecmd("ex t:[12,22]", batch=True)
    bitmap1 = fs.read_bitmap()

    shell.onecmd("create /allocate:5 t:[10,20]test.asc", batch=True)
    l = list(fs.filter_entries_list("[10,20]*.asc"))
    assert len(l) == 4
    bitmap2 = fs.read_bitmap()
    assert bitmap2.used() == bitmap1.used() + 5

    shell.onecmd("create /allocate:10 t:[10,20]test.asc", batch=True)
    bitmap3 = fs.read_bitmap()
    assert bitmap3.used() == bitmap1.used() + 10

    shell.onecmd("create /allocate:5 t:[10,20]test.asc", batch=True)
    bitmap4 = fs.read_bitmap()
    assert bitmap4.used() == bitmap2.used()
    assert bitmap4 == bitmap2

    shell.onecmd("delete t:[10,20]test.asc", batch=True)
    bitmap5 = fs.read_bitmap()
    l = list(fs.filter_entries_list("[10,20]*.asc"))
    assert len(l) == 3
    assert bitmap5.used() == bitmap1.used()
    assert bitmap5 == bitmap1

    shell.onecmd("create/directory t:[5,5]", batch=True)
    bitmap6 = fs.read_bitmap()
    assert bitmap6.used() > bitmap1.used()

    shell.onecmd("create /allocate:5 t:[5,5]test.pal", batch=True)

    shell.onecmd("delete t:[5,5]", batch=True)
    bitmap7 = fs.read_bitmap()
    assert bitmap7.used() == bitmap1.used()
    assert bitmap7 == bitmap1

    x = fs.read_bytes("[11,21]M50.asc")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.filter_entries_list("[11,21]*.asc"))
    assert len(l) == 3


def test_tss8_write_file():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}.mo", batch=True)
    shell.onecmd(f"mount t: /tss8 {DSK}.mo", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, TSS8Filesystem)

    blocks = 5
    filename = "[10,20]data.asc"
    data = b""
    for i in range(0, blocks):
        data += chr(i + 65).encode("ascii") * TSS8_BLOCK_SIZE_BYTES
    fs.write_bytes(filename, data)

    data_read = fs.read_bytes(filename)
    assert data_read == data

    f = fs.open_file(filename)
    for i in range(0, blocks):
        block_data = f.read_block(i)
        assert block_data == chr(i + 65).encode("ascii") * TSS8_BLOCK_SIZE_BYTES
    f.close()

    f = fs.open_file(filename)
    for i in range(0, blocks):
        tmp = chr(i + 85).encode("ascii") * TSS8_BLOCK_SIZE_BYTES
        f.write_block(tmp, i)
    f.close()

    f = fs.open_file(filename)
    for i in range(0, blocks):
        block_data = f.read_block(i)
        assert block_data == chr(i + 85).encode("ascii") * TSS8_BLOCK_SIZE_BYTES
    f.close()

    f = fs.open_file(filename)
    for i in range(0, blocks // 2):
        tmp = chr(i + 85).encode("ascii") * TSS8_BLOCK_SIZE_BYTES * 2
        f.write_block(tmp, i * 2, number_of_blocks=2)
    f.close()

    f = fs.open_file(filename)
    for i in range(0, blocks // 2):
        block_data = f.read_block(i * 2, number_of_blocks=2)
        assert block_data == chr(i + 85).encode("ascii") * TSS8_BLOCK_SIZE_BYTES * 2
    f.close()



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import struct
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile
from ..commons import BLOCK_SIZE, READ_FILE_FULL, dump_struct, filename_match
from ..device.abstract import AbstractDevice
from .abstract import AbstractAppleDiskFilesystem

__all__ = [
    "PascalFile",
    "PascalDirectoryEntry",
    "PascalFilesystem",
]

# Apple II Pascal Filesystem
# https://archive.org/details/apple-ii-pascal-1.3/page/n803/mode/2up

FILENAME_LEN = 15  # max filename length
DIR_BLOCK = 2  # directory first block
DIR_SIZE = 4  # 4 blocks
MAX_DIR_ENTRIES = 77  # max directory entries
VOLUME_DIRECTORY_ENTRY_FORMAT = "<HHH8sHHHH4s"
VOLUME_DIRECTORY_ENTRY_SIZE = struct.calcsize(VOLUME_DIRECTORY_ENTRY_FORMAT)
DIRECTORY_ENTRY_FORMAT = "<HHH16sHH"
DIRECTORY_ENTRY_SIZE = struct.calcsize(DIRECTORY_ENTRY_FORMAT)
DEFAULT_VOLUME_NAME = "PASCAL"

# Diskette file type, Pag 34
# https://archive.org/details/Apple_Pascal_Operating_System_Reference_Manual_HQ/page/n33/mode/2up

FILE_TYPE_UNTYPED = 0  # Untyped file
FILE_TYPE_BAD = 1  # Bad block
FILE_TYPE_CODE = 2  # Machine-executable code
FILE_TYPE_TEXT = 3  # Human-readable text file
FILE_TYPE_INFO = 4  # (not used)
FILE_TYPE_DATA = 5  # Data file
FILE_TYPE_GRAF = 6  # (not used)
FILE_TYPE_FOTO = 7  # (not used)
FILE_TYPE_SECUREDIR = 8  # (not used)

FILE_TYPES = {
    FILE_TYPE_BAD: "BAD",
    FILE_TYPE_CODE: "CODE",
    FILE_TYPE_TEXT: "TEXT",
    FILE_TYPE_INFO: "INFO",
    FILE_TYPE_DATA: "DATA",
    FILE_TYPE_GRAF: "GRAF",
    FILE_TYPE_FOTO: "FOTO",
}


def pascal_to_str(buffer: bytes) -> str:
    """
    Convert a Pascal string to a Python string
    """
    length = buffer[0]
    return buffer[1 : length + 1].decode("ascii", errors="ignore")


def str_to_pascal(val: str) -> bytes:
    """
    Convert a Python string to a Pascal string
    """
    length = len(val)
    return bytes([length]) + val.encode("ascii", errors="ignore")


def pascal_to_date(val: int) -> t.Optional[date]:
    """
    Translate Pascal date to Python date
    """
    if val == 0:
        return None
    year = (val >> 9) & 0x7F
    day = (val >> 4) & 0x1F
    month = val & 0x0F
    year = year + 1900 if year >= 80 else year + 2000
    try:
        return date(year, month, day)
    except:
        return None


def date_to_pascal(val: t.Optional[date]) -> int:
    """
    Translate Python date to Pascal date
    """
    if val is None:
        return 0
    year = val.year
    year = year - 2000 if year >= 2000 else year - 1900
    return (year << 9) | (val.day << 4) | val.month


def pascal_canonical_filename(fullname: t.Optional[str], wildcard: bool = False) -> t.Optional[str]:
    """
    Generate the canonical Pascal filename
    """
    if fullname:
        fullname = fullname[:FILENAME_LEN].upper()
    return fullname


def pascal_get_raw_file_type(file_type: t.Optional[str], default: int = FILE_TYPE_TEXT) -> int:
    """
    Get the file type id from a string
    """
    if not file_type:
        return default
    file_type = file_type.upper()
    for file_id, file_str in FILE_TYPES.items():
        if file_str == file_type:
            return file_id
    raise Exception("?KMON-F-Invalid file type specified with option")


def format_long_filetype(file_type: int) -> str:
    """
    Format file type
    """
    if file_type == FILE_TYPE_BAD:
        return "Bad disk"
    elif file_type == FILE_TYPE_CODE:
        return "Codefile"
    elif file_type == FILE_TYPE_TEXT:
        return "Textfile"
    elif file_type == FILE_TYPE_INFO:
        return "Infofile"
    elif file_type == FILE_TYPE_DATA:
        return "Datafile"
    elif file_type == FILE_TYPE_GRAF:
        return "Graffile"
    elif file_type == FILE_TYPE_FOTO:
        return "Fotofile"
    else:
        return "ILLEGAL"


class PascalFile(AbstractFile):
    entry: "PascalDirectoryEntry"
    closed: bool

    def __init__(self, entry: "PascalDirectoryEntry"):
        self.entry = entry
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if self.closed or block_number < 0 or number_of_blocks < 0:
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        if block_number + number_of_blocks > self.entry.length:
            number_of_blocks = self.entry.length - block_number
        return self.entry.fs.read_block(
            self.entry.start_block + block_number,
            number_of_blocks,
        )

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        self.entry.fs.write_block(
            buffer,
            self.entry.start_block + block_number,
            number_of_blocks,
        )

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class VolumeDirectory:

    fs: "PascalFilesystem"
    start_block: int  # Start block
    following_block: int  # Last block + 1
    raw_file_type: int = 0  # File type
    volume_name: str  # Volume name
    number_of_blocks: int  # Number of blocks
    number_of_files: int  # Number of files
    last_access_time: int  # Last access time
    raw_most_recently_date: int  # Most recently date setting
    directory_entries: t.List["PascalDirectoryEntry"]  # 77 entries, even if empty/unused

    def __init__(self, fs: "PascalFilesystem"):
        self.fs = fs

    @classmethod
    def read(cls, fs: "PascalFilesystem") -> "VolumeDirectory":
        """
        Read the volume directory from disk
        """
        self = cls(fs)
        buffer = fs.read_block(DIR_BLOCK, number_of_blocks=DIR_SIZE)
        (
            self.start_block,
            self.following_block,
            self.raw_file_type,
            volume_name,
            self.number_of_blocks,
            self.number_of_files,
            self.last_access_time,
            self.raw_most_recently_date,
            _,
        ) = struct.unpack_from(VOLUME_DIRECTORY_ENTRY_FORMAT, buffer, 0)
        self.volume_name = pascal_to_str(volume_name)
        # Read directory entries
        self.directory_entries = [
            PascalDirectoryEntry.read(self.fs, buffer, i * DIRECTORY_ENTRY_SIZE) for i in range(1, MAX_DIR_ENTRIES + 1)
        ]
        return self

    def write(self) -> None:
        """ "
        Write the volume directory to disk
        """
        buffer = bytearray(BLOCK_SIZE * DIR_SIZE)
        # Update number of files
        entries = [entry for entry in self.iterdir() if entry.filename]
        self.number_of_files = len(entries)
        # Write volume directory entry
        struct.pack_into(
            VOLUME_DIRECTORY_ENTRY_FORMAT,
            buffer,
            0,
            self.start_block,
            self.following_block,
            self.raw_file_type,
            str_to_pascal(self.volume_name),
            self.number_of_blocks,
            self.number_of_files,
            self.last_access_time,
            self.raw_most_recently_date,
            b"\0" * 4,
        )
        # Write directory entries
        for i, entry in enumerate(sorted(entries), start=1):
            struct.pack_into(
                DIRECTORY_ENTRY_FORMAT,
                buffer,
                i * DIRECTORY_ENTRY_SIZE,
                entry.start_block,
                entry.following_block,
                entry.raw_file_type,
                str_to_pascal(entry.filename),
                entry.last_block_bytes,
                entry.raw_mod_date,
            )
        self.directory_entries = entries + [
            PascalDirectoryEntry(self.fs) for _ in range(MAX_DIR_ENTRIES - len(entries))
        ]
        self.fs.write_block(bytes(buffer), DIR_BLOCK, number_of_blocks=DIR_SIZE)

    @property
    def most_recently_date(self) -> t.Optional[date]:
        return pascal_to_date(self.raw_most_recently_date)

    def iterdir(self, include_empty_area: bool = False) -> t.Iterator["PascalDirectoryEntry"]:
        """
        Iterate over directory entries
        """
        count = self.number_of_files
        following_block = self.following_block
        for entry in self.directory_entries:
            if count == 0:
                break
            if include_empty_area:
                # Add empty area between files
                free_blocks = entry.start_block - following_block
                if free_blocks > 0:
                    free_entry = PascalDirectoryEntry(
                        self.fs,
                        start_block=following_block,
                        following_block=entry.start_block,
                    )
                    yield free_entry
            if entry.filename:
                count -= 1
                yield entry
            following_block = entry.following_block

        if include_empty_area:
            # Add empty area at the end
            free_blocks = self.number_of_blocks - following_block
            if free_blocks > 0:
                free_entry = PascalDirectoryEntry(
                    self.fs,
                    start_block=following_block,
                    following_block=self.number_of_blocks,
                )
                yield free_entry

    def allocate_space(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,  # optional file type
        last_block_bytes: int = 0,  # number of bytes in last block
    ) -> "PascalDirectoryEntry":
        """
        Allocate space for a new file
        """
        if self.number_of_files >= MAX_DIR_ENTRIES:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        # Search for empty space
        for entry in self.iterdir(include_empty_area=True):
            # Check if there is enough space
            if entry.is_empty and entry.length >= number_of_blocks:
                # Create the new entry
                entry.filename = fullname
                entry.last_block_bytes = last_block_bytes
                entry.raw_file_type = pascal_get_raw_file_type(file_type)
                entry.raw_mod_date = date_to_pascal(creation_date or date.today())
                entry.following_block = entry.start_block + number_of_blocks
                # Update volume directory
                self.directory_entries = self.directory_entries + [entry]
                self.number_of_files += 1
                return entry
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def __str__(self) -> str:
        try:
            date_str = self.most_recently_date.strftime("%d-%b-%y").lstrip("0")  # type: ignore
        except Exception:
            date_str = ""
        return f"Volume: {self.volume_name}  Blocks: {self.number_of_blocks}  Files: {self.number_of_files}  Access time: {self.last_access_time}  Set date: {date_str}"


class PascalDirectoryEntry(AbstractDirectoryEntry):

    fs: "PascalFilesystem"
    start_block: int  # Start block
    following_block: int  # Last block + 1
    filename: str  # Filename
    last_block_bytes: int  # Number of bytes used in last block
    raw_file_type: int  # File type
    raw_mod_date: int  # Modification date

    def __init__(
        self,
        fs: "PascalFilesystem",
        start_block: int = 0,
        following_block: int = 0,
        filename: str = "",
        last_block_bytes: int = 0,
        raw_file_type: int = 0,
        raw_mod_date: int = 0,
    ):
        self.fs = fs
        self.start_block = start_block
        self.following_block = following_block
        self.filename = filename
        self.last_block_bytes = last_block_bytes
        self.raw_file_type = raw_file_type
        self.raw_mod_date = raw_mod_date

    @classmethod
    def read(cls, fs: "PascalFilesystem", buffer: bytes, position: int) -> "PascalDirectoryEntry":
        self = cls(fs)
        (
            self.start_block,  # start block
            self.following_block,  # last block + 1
            self.raw_file_type,  # file type
            filename,  # filename
            self.last_block_bytes,  # number of bytes in last block
            self.raw_mod_date,  # modification date
        ) = struct.unpack_from(DIRECTORY_ENTRY_FORMAT, buffer, position)
        self.filename = pascal_to_str(filename)
        return self

    @property
    def length(self) -> int:
        """
        Length in blocks
        """
        return self.following_block - self.start_block

    @property
    def is_empty(self) -> bool:
        return self.length == 0 or not self.filename

    @property
    def file_type(self) -> t.Optional[str]:
        """
        File type (e.g. DATA)
        """
        return FILE_TYPES.get(self.raw_file_type)

    @property
    def long_file_type(self) -> str:
        """
        File type long description (e.g. Datafile)
        """
        return format_long_filetype(self.raw_file_type)

    @property
    def fullname(self) -> str:
        return self.filename

    @property
    def basename(self) -> str:
        return self.filename

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return (self.length - 1) * self.get_block_size() + self.last_block_bytes

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    @property
    def creation_date(self) -> t.Optional[date]:
        return pascal_to_date(self.raw_mod_date)

    def delete(self) -> bool:
        """
        Delete the file
        """
        volume_dir = VolumeDirectory.read(self.fs)
        for entry in volume_dir.directory_entries:
            if entry.start_block == self.start_block and entry.filename == self.filename:
                self.start_block = entry.start_block = 0
                self.following_block = entry.following_block = 0
                self.filename = entry.filename = ""
                self.last_block_bytes = entry.last_block_bytes = 0
                self.raw_file_type = entry.raw_file_type = 0
                self.raw_mod_date = entry.raw_mod_date = 0
                volume_dir.write()
                return True
        return False

    def write(self) -> bool:
        """
        Write the directory entry
        """
        volume_dir = VolumeDirectory.read(self.fs)
        for entry in volume_dir.directory_entries:
            if entry.start_block == self.start_block:
                volume_dir.write()
                return True
        return False

    def open(self, file_mode: t.Optional[str] = None) -> PascalFile:
        """
        Open a file
        """
        return PascalFile(self)

    def __lt__(self, other: "PascalDirectoryEntry") -> bool:
        return self.start_block < other.start_block

    def __gt__(self, other: "PascalDirectoryEntry") -> bool:
        return self.start_block > other.start_block

    def __str__(self) -> str:
        date = self.creation_date and self.creation_date.strftime("%d-%b-%y") or ""
        if self.creation_date and not self.is_empty:
            date = self.creation_date.strftime("%d-%b-%y").lstrip("0")
        else:
            date = ""
        fullname = self.fullname if self.fullname else "< UNUSED >"
        return f"{fullname:<15} {self.length:>6}  {date:>9} {self.start_block:>4} -> {self.following_block:>4}  {self.last_block_bytes:>3}  {self.raw_file_type} {self.long_file_type}"

    def __repr__(self) -> str:
        return str(self)


class PascalFilesystem(AbstractAppleDiskFilesystem):
    """
    Apple II Pascal Filesystem
    """

    fs_name = "pascal"
    fs_description = "Apple II Pascal"
    fs_platforms = ["apple2"]

    volume_name: str = ""  # Volume name

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        **kwargs: t.Union[bool, str],
    ) -> "PascalFilesystem":
        self = cls(file_or_dev)
        # Read volume dir
        volume_dir = VolumeDirectory.read(self)
        if not volume_dir.volume_name:
            self.dev.prodos_order = True
            volume_dir = VolumeDirectory.read(self)
            if not volume_dir.volume_name:
                raise OSError(errno.EIO, os.strerror(errno.EIO))
        self.volume_name = volume_dir.volume_name
        return self

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["PascalDirectoryEntry"]:
        if pattern:
            pattern = pascal_canonical_filename(pattern, wildcard=wildcard)
        volume_dir = VolumeDirectory.read(self)
        for entry in volume_dir.iterdir(include_empty_area=include_all):
            if filename_match(entry.basename, pattern, wildcard):
                yield entry

    @property
    def entries_list(self) -> t.Iterator["PascalDirectoryEntry"]:
        volume_dir = VolumeDirectory.read(self)
        yield from volume_dir.iterdir()

    def get_file_entry(self, fullname: str) -> PascalDirectoryEntry:
        fullname = pascal_canonical_filename(fullname)  # type: ignore
        for entry in self.entries_list:
            if entry.fullname == fullname and not entry.is_empty:
                return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def read_bytes(self, fullname: str, file_type: t.Optional[str] = None) -> bytes:  # fullname=filename+ext
        entry = self.get_file_entry(fullname)
        return self.read_block(entry.start_block, entry.length)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        """
        Write content to a file
        """
        number_of_blocks = int(math.ceil(len(content) / BLOCK_SIZE))
        last_block_bytes = len(content) % BLOCK_SIZE
        entry = self.create_file(
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
            file_type=file_type,
            last_block_bytes=last_block_bytes,
        )
        if entry is not None:
            content = content + (b"\0" * BLOCK_SIZE)  # pad with zeros
            f = entry.open(file_mode)
            try:
                f.write_block(content, block_number=0, number_of_blocks=number_of_blocks)
            finally:
                f.close()

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,  # optional file type
        last_block_bytes: int = 0,  # number of bytes in last block
    ) -> t.Optional[PascalDirectoryEntry]:
        fullname = pascal_canonical_filename(fullname)  # type: ignore
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        volume_dir = VolumeDirectory.read(self)
        entry = volume_dir.allocate_space(
            fullname=fullname,
            number_of_blocks=number_of_blocks,
            creation_date=creation_date,
            file_type=file_type,
            last_block_bytes=last_block_bytes,
        )
        volume_dir.write()
        return entry

    def chdir(self, fullname: str) -> bool:
        return False

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        volume_dir = VolumeDirectory.read(self)
        files = 0  # Number of listed files
        blocks = 0  # Used blocks
        unused = 0  # Unused blocks
        largest_unused = 0  # Largest unused block
        if not options.get("brief"):
            sys.stdout.write(f"{volume_id}:\n")
        if pattern:
            pattern = pascal_canonical_filename(pattern, wildcard=True)
        for x in volume_dir.iterdir(include_empty_area=True):
            if options.get("brief"):
                if filename_match(x.basename, pattern, wildcard=True):
                    sys.stdout.write(f"{x.fullname}\n")
            else:
                if x.is_empty:
                    if x.length > largest_unused:
                        largest_unused = x.length
                    unused = unused + x.length
                    if not pattern:
                        sys.stdout.write(f"{'< UNUSED >':<15} {x.length:>6}  {'':>9} {x.start_block:>4}\n")
                else:
                    blocks = blocks + x.length
                    date = x.creation_date and x.creation_date.strftime("%d-%b-%y") or ""
                    if x.creation_date:
                        date = x.creation_date.strftime("%d-%b-%y").lstrip("0")
                    else:
                        date = ""
                    if filename_match(x.basename, pattern, wildcard=True):
                        files = files + 1
                        sys.stdout.write(
                            f"{x.fullname:<15} {x.length:>6}  {date:>9} {x.start_block:>4}  {x.last_block_bytes:>3}  {x.long_file_type}\n"
                        )
        if not options.get("brief"):
            sys.stdout.write(
                f"{files}/{volume_dir.number_of_files} files <listed/in dir>, {blocks} blocks used, {unused} unused, {largest_unused} in largest\n"
            )

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if arg:
            # Dump by path
            entry = self.get_file_entry(arg)
            entry_dict = dict(entry.__dict__)
            del entry_dict["fs"]
            sys.stdout.write(dump_struct(entry_dict) + "\n")
        else:
            # Dump the entire filesystem
            volume_dir = VolumeDirectory.read(self)
            volume_dir_dict = dict(volume_dir.__dict__)
            del volume_dir_dict["fs"]
            del volume_dir_dict["directory_entries"]
            sys.stdout.write(dump_struct(volume_dir_dict))
            sys.stdout.write("\n\n")
            sys.stdout.write("Nr  Filename        Blocks  Date     Start     End Size  File type\n")
            sys.stdout.write("--  --------        ------  ----     -----     --- ----  ---------\n")
            for i, entry in enumerate(volume_dir.directory_entries, start=1):
                # Skip null entries
                if (
                    entry.filename
                    or entry.length
                    or entry.raw_file_type
                    or entry.start_block
                    or entry.following_block
                    or entry.last_block_bytes
                    or entry.raw_mod_date
                    or options.get("full")
                ):
                    sys.stdout.write(f"{i:>2}# {entry}\n")

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    @classmethod
    def initialize(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        **kwargs: t.Union[bool, str],
    ) -> "PascalFilesystem":
        """
        Initialize the filesystem
        """
        self = cls(file_or_dev)
        try:
            volume_name = kwargs["name"].strip().upper() or DEFAULT_VOLUME_NAME  # type: ignore
        except Exception:
            volume_name = DEFAULT_VOLUME_NAME
        volume_dir = VolumeDirectory(self)
        volume_dir.start_block = 0
        volume_dir.following_block = DIR_BLOCK + DIR_SIZE
        volume_dir.raw_file_type = 0
        volume_dir.volume_name = volume_name
        volume_dir.number_of_blocks = self.dev.get_size() // BLOCK_SIZE
        volume_dir.number_of_files = 0
        volume_dir.last_access_time = 0
        volume_dir.raw_most_recently_date = date_to_pascal(date.today())
        volume_dir.directory_entries = [PascalDirectoryEntry(self, 0, 0) for _ in range(MAX_DIR_ENTRIES)]
        volume_dir.write()
        self.volume_name = volume_name
        return self

    def get_pwd(self) -> str:
        return ""

    def get_types(self) -> t.List[str]:
        """
        Get the list of the supported file types
        """
        return list(FILE_TYPES.values())



import random
from datetime import date, timedelta

import pytest

from xferx.apple2.pascalfs import PascalFilesystem, date_to_pascal, pascal_to_date
from xferx.shell import Shell

DSK = "tests/dsk/pascal.dsk"


def test_date_round_trip():
    start_year = 1980
    end_year = 2020
    for i in range(0, 100):
        start_date = date(start_year, 1, 1)
        end_date = date(end_year, 12, 31)
        random_seconds = random.randint(0, int((end_date - start_date).total_seconds()))
        random_seconds = random_seconds - (random_seconds % 60)
        dt = start_date + timedelta(seconds=random_seconds)
        assert pascal_to_date(date_to_pascal(dt)) == dt


def test_pascal():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /pascal {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, PascalFilesystem)

    shell.onecmd("dir t:", batch=True)
    shell.onecmd("dir/brief t:", batch=True)
    shell.onecmd("dir/brief t:*.txt", batch=True)
    shell.onecmd("dir/brief t:*.notfound", batch=True)
    shell.onecmd("type t:1.txt", batch=True)

    x = fs.read_bytes("50.txt")
    x = x.rstrip(b"\0")
    assert len(x) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.filter_entries_list("*.TXT"))
    assert len(l) == 10
    for x in l:
        assert not x.is_empty
        assert str(x)


def test_pascal_init():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /pascal {DSK}", batch=True)
    shell.onecmd(f"create /allocate:280 {DSK}.mo", batch=True)
    shell.onecmd(f"init /pascal {DSK}.mo", batch=True)
    shell.onecmd(f"mount ou: /pascal {DSK}.mo", batch=True)
    shell.onecmd("dir ou:", batch=True)
    shell.onecmd("copy t:*.txt ou:", batch=True)
    fs = shell.volumes.get('OU')

    x1 = fs.read_bytes("50.txt")
    x1 = x1.rstrip(b"\0")
    assert len(x1) == 2200
    for i in range(0, 50):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x1

    with pytest.raises(Exception):
        shell.onecmd("delete ou:aaa", batch=True)
    shell.onecmd("delete ou:50.txt", batch=True)
    assert fs.read_bytes("10.txt")
    with pytest.raises(FileNotFoundError):
        fs.read_bytes("50.txt")

    # Test init mounted volume
    shell.onecmd("init ou:", batch=True)
    with pytest.raises(Exception):
        print(fs.read_bytes("10.tx"))



import pytest

from xferx.pdp11.files11fs import HOME_BLOCK, INDEXF_SYS, Files11File, Files11Filesystem
from xferx.shell import Shell

DSK = "tests/dsk/files11.dsk"


def test_files11():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount t: /files11 {DSK}", batch=True)
    fs = shell.volumes.get('T')
    assert isinstance(fs, Files11Filesystem)

    indexfs = fs.read_file_header(INDEXF_SYS)
    f = Files11File(indexfs)

    bootstrap_block = fs.read_block(0)
    assert f.read_block(0) == bootstrap_block
    #
    home = fs.read_block(HOME_BLOCK)
    assert b"DECFILE11A" in home
    assert f.read_block(HOME_BLOCK) == home
    f.close()

    with pytest.raises(Exception):
        [x.basename for x in fs.filter_entries_list(pattern="[33,44]*.*")]

    assert fs.chdir("[0,0]")
    files = [x.basename for x in fs.filter_entries_list(pattern="*.*")]
    assert "INDEXF.SYS" in files
    assert "BITMAP.SYS" in files
    assert "000000.DIR" in files
    files = [x.fullname for x in fs.filter_entries_list(pattern="*.*")]
    assert "[0,0]INDEXF.SYS" in files
    assert "[0,0]BITMAP.SYS" in files
    assert "[0,0]000000.DIR" in files



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import os
import struct
import typing as t
from datetime import date, datetime

from ..commons import BLOCK_SIZE
from .pascalfs import pascal_to_str, str_to_pascal
from .prodosfs import (
    DEFAULT_ACCESS,
    PAS_FILE_TYPE,
    PASCAL_AREA_STORAGE_TYPE,
    AbstractDirectoryFileEntry,
    FileEntry,
    ProDOSAbstractDirEntry,
    ProDOSBitmap,
    ProDOSFilesystem,
    RegularFileEntry,
    date_to_prodos,
    parse_file_aux_type,
)

PPM_HEADER_BLOCKS = 2  # Pascal Volume Header length in blocks
PPM_HEADER_FORMAT = "<HH4s"
PPM_MAX_VOLUMES = 31
PPM_INFO_FORMAT = "<HHBBH"  # Volume info
PPM_INFO_SIZE = struct.calcsize(PPM_INFO_FORMAT)
PPM_DESCRIPTION_OFFSET = 0x100
PPM_DESCRIPTION_LENGTH = 16  # bytes
PPM_NAME_OFFSET = 0x300
PPM_NAME_LENGTH = 8  # bytes

__all__ = [
    "PPMDirectoryEntry",
    "PPMVolumeEntry",
]


class PPMDirectoryEntry(AbstractDirectoryFileEntry):
    """
    Pascal ProFile Manager (PPM) Partition
    https://ciderpress2.com/formatdoc/PPM-notes.html

    Pascal ProFile Manager Manual
    https://archive.org/details/a2ppfmm/

    The Pascal Partition is a contiguous file that starts
    at key_pointer and extends to the end of the disk.
    It is created by the Apple Pascal ProFile Manager,
    and it has file type PAS and storage type 4.

    Internally, the Pascal Partition is divided into 1-31 volumes.

      Length                                Offset
             +----------------------------+
     2 block |   Pascal Volume Directory  | key_pointer
             |                            |
             |----------------------------|
             |      Pascal Volume 1       |
             |                            |
             |----------------------------|
             |      Pascal Volume 2       |
             |                            |
             |----------------------------|
             | ...                        |
             |                            | disk last block
             +----------------------------+

    Pascal Volume Directory

       Field                                Byte of
      Length                                Header
             +----------------------------+
     2 byte  |       Size in blocks       | $00
             |                            | $01
             +----------------------------+
     2 byte  |     Number of volumes      | $02
             |                            | $03
             +----------------------------+
     4 byte  |  'PPM' as Pascal string    | $04
             |                            | $07
             |----------------------------|
             /                            /
             |----------------------------| Volume n info
     2 byte  |   Volume n start block     | $00 + n * $8
             |                            |
             |----------------------------|
     2 byte  |      Volume n length       | $02 + n * $8
             |         in blocks          |
             |----------------------------|
     1 byte  |   Volume n default unit    | $04 + n * $8
             |----------------------------|
     1 byte  | Volume n write protection  | $05 + n * $8
             |----------------------------|
     2 byte  |          Reserved          | $06 + n * $8
             |                            |
             |----------------------------|
             /                            /
             |----------------------------|
     16 byte |   Volume n description as  | $100 + n * $16
             |       Pascal string        |
             |----------------------------|
             /                            /
             |----------------------------|
      8 byte |     Volume n name as       | $300 + n * $8
             |       Pascal string        |
             |----------------------------|

    """

    volumes: int = 0  # Allocated volumes (up to 31)
    ppm_name: str = ""  # 'PPM' as Pascal string

    @classmethod
    def read(
        cls, fs: "ProDOSFilesystem", parent: t.Optional["FileEntry"], buffer: bytes, position: int = 0
    ) -> "FileEntry":
        self: PPMDirectoryEntry = super().read(fs, parent, buffer, position)  # type: ignore
        # Read the header
        buffer = self.fs.read_block(self.key_pointer, number_of_blocks=PPM_HEADER_BLOCKS)
        (
            _,  # size in blocks
            self.volumes,  # number of allocated volumes
            ppm_name,  # 'PPM' as Pascal string
        ) = struct.unpack_from(PPM_HEADER_FORMAT, buffer, 0)
        self.ppm_name = pascal_to_str(ppm_name)
        return self

    @classmethod
    def create(
        cls,
        fs: "ProDOSFilesystem",
        parent: t.Optional["AbstractDirectoryFileEntry"],
        filename: str,
        length: int,  # Length in blocks
        bitmap: "ProDOSBitmap",
        creation_date: t.Optional[t.Union[date, datetime]] = None,  # optional creation date
        access: int = DEFAULT_ACCESS,  # optional access
        file_type: t.Optional[str] = None,  # optional file type
        aux_type: int = 0,  # optional aux type
        length_bytes: t.Optional[int] = None,  # optional length int bytes
        resource_length_bytes: t.Optional[int] = None,  # not used
    ) -> "PPMDirectoryEntry":
        """
        Create a new Pascal ProFile Manager (PPM) Partition
        """
        # Allocate the contiguous blocks
        last_block = fs.get_size() // BLOCK_SIZE - 1
        first_block = last_block - length
        for block_number in range(first_block, last_block + 1):
            if not bitmap.is_free(block_number):
                raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
            bitmap.set_used(block_number)
        # Create the entry
        self = cls(fs, parent)
        self.ppm_name = "PPM"
        self.volumes = 0
        self.storage_type = PASCAL_AREA_STORAGE_TYPE
        self.blocks_used = length
        self.length = length_bytes if length_bytes is not None else length * BLOCK_SIZE
        self.filename = filename
        self.key_pointer = first_block
        self.access = access
        if isinstance(creation_date, datetime):
            self.last_mod_date = creation_date
        elif isinstance(creation_date, date):
            self.last_mod_date = datetime.combine(creation_date, datetime.min.time())
        else:
            self.last_mod_date = datetime.now()
        self.raw_creation_date = date_to_prodos(self.last_mod_date)
        self.prodos_file_type, self.aux_type = parse_file_aux_type(
            file_type, default=PAS_FILE_TYPE, default_aux_type=aux_type
        )
        # Write the volume header
        buffer = bytearray(self.fs.read_block(self.key_pointer, number_of_blocks=2))
        struct.pack_into(
            PPM_HEADER_FORMAT,
            buffer,
            0,
            self.blocks_used,
            0,  # volumes
            self.ppm_name.encode("ascii"),
        )
        self.fs.write_block(buffer, self.key_pointer, number_of_blocks=2)
        # Write the entry
        if parent is not None:
            try:
                parent.update_dir_entry(self, create=True)
            except OSError:
                # Directory is full, grow it
                parent.grow(bitmap)
                parent.update_dir_entry(self, create=True)
        return self

    def update_dir_entry(
        self,
        entry: "ProDOSAbstractDirEntry",
        entry_class: t.Type["ProDOSAbstractDirEntry"] = FileEntry,
        create: bool = False,
        delete: bool = False,
    ) -> t.Optional[t.Tuple[int, int]]:
        """
        Update/create/delete a directory entry
        """
        volume_entry: PPMVolumeEntry = entry  # type: ignore
        if create:
            return self.create_volume_entry(volume_entry)
        elif delete:
            return self.delete_volume_entry(volume_entry)
        else:
            return self.update_volume_entry(volume_entry)

    def create_volume_entry(self, volume_entry: "PPMVolumeEntry") -> t.Optional[t.Tuple[int, int]]:
        """
        Create a volume entry in the Pascal Volume Directory
        """
        start = self.key_pointer + PPM_HEADER_BLOCKS
        position = start
        volume_number = 0
        volumes: list[PPMVolumeEntry] = list(self.iterdir())
        for volume in volumes:
            if volume.key_pointer - position >= volume_entry.blocks_used:
                break
            position = volume.key_pointer + volume.blocks_used
        volume_number += 1
        if volume_number > PPM_MAX_VOLUMES:
            raise OSError(errno.ENOSPC, "Too many volumes")
        if position + volume_entry.blocks_used >= self.fs.get_size() // BLOCK_SIZE:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        volume_entry.key_pointer = position
        volume_entry.volume_number = volume_number
        volumes.append(volume_entry)
        self.write_pascal_volume_directory(volumes)
        return (self.key_pointer, volume_number)

    def delete_volume_entry(self, volume_entry: "PPMVolumeEntry") -> t.Optional[t.Tuple[int, int]]:
        """
        Delete a volume entry from the Pascal Volume Directory
        """
        volumes: list[PPMVolumeEntry] = []
        volume_number = 0
        for i, volume in enumerate(self.iterdir(), start=1):
            if volume.key_pointer != volume_entry.key_pointer:
                volumes.append(volume)
            else:
                volume_number = i
        if volume_number == 0:
            return None
        else:
            self.write_pascal_volume_directory(volumes)
            return (self.key_pointer, volume_number)

    def update_volume_entry(self, volume_entry: "PPMVolumeEntry") -> t.Optional[t.Tuple[int, int]]:
        """
        Update a volume entry in the Pascal Volume Directory
        """
        volumes: list[PPMVolumeEntry] = []
        volume_number = 0
        for i, volume in enumerate(self.iterdir(), start=1):
            if volume.key_pointer == volume_entry.key_pointer:
                volumes.append(volume_entry)
                volume_number = i
            else:
                volumes.append(volume)
        if volume_number == 0:
            return None
        else:
            self.write_pascal_volume_directory(volumes)
            return (self.key_pointer, volume_number)

    def write_pascal_volume_directory(self, volumes: t.List["PPMVolumeEntry"]) -> None:
        """
        Write the Pascal Volume Directory to the disk
        """
        volumes = sorted(volumes, key=lambda x: x.key_pointer)
        self.volumes = len(volumes)
        buffer = bytearray(self.fs.read_block(self.key_pointer, number_of_blocks=PPM_HEADER_BLOCKS))
        # Write the header
        struct.pack_into(
            PPM_HEADER_FORMAT,
            buffer,
            0,
            self.blocks_used,
            self.volumes,
            self.ppm_name.encode("ascii"),
        )
        # Write the volumes
        for volume_number, volume in enumerate(volumes, start=1):
            position = volume_number * PPM_INFO_SIZE
            volume.write_buffer(buffer, position)
        self.fs.write_block(bytes(buffer), self.key_pointer, number_of_blocks=PPM_HEADER_BLOCKS)

    def blocks(self, include_indexes: bool = False) -> t.Iterator[int]:
        """
        Iterate over the blocks used by the PPM Partition
        """
        yield from range(self.key_pointer, self.fs.get_size() // BLOCK_SIZE)

    def iterdir(self) -> t.Iterator["PPMVolumeEntry"]:
        """
        Read the volumes from the Pascal Volume Directory
        """
        buffer = self.fs.read_block(self.key_pointer, number_of_blocks=PPM_HEADER_BLOCKS)
        for volume_number in range(1, self.volumes + 1):
            yield PPMVolumeEntry.read(self.fs, self, buffer, volume_number * PPM_INFO_SIZE)

    def grow(self, bitmap: "ProDOSBitmap") -> None:
        """
        Grow the directory (not supported by PPM))
        """
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))


class PPMVolumeEntry(RegularFileEntry):
    """
    Volume (a contiguous area) in the
    Pascal ProFile Manager (PPM) Partition
    """

    volume_number: int = 0  # Volume number (1-31)
    volume_name: str = ""  # Volume name (8 characters)
    description: str = ""  # Volume description (15 characters)
    default_unit: int = 0  # Default unit number
    write_protection: int = 0  # Write protection flag

    @classmethod
    def read(
        cls,
        fs: "ProDOSFilesystem",
        parent: t.Optional["FileEntry"],
        buffer: bytes,
        position: int = 0,
    ) -> "PPMVolumeEntry":
        assert parent is not None
        self = PPMVolumeEntry(fs, parent)
        self.volume_number = position // PPM_INFO_SIZE
        (
            self.key_pointer,
            self.blocks_used,
            self.default_unit,
            self.write_protection,
            _,
        ) = struct.unpack_from(PPM_INFO_FORMAT, buffer, self.volume_number * PPM_INFO_SIZE)
        description_position = PPM_DESCRIPTION_OFFSET + self.volume_number * PPM_DESCRIPTION_LENGTH
        self.description = pascal_to_str(buffer[description_position : description_position + 8])
        volume_name_position = PPM_NAME_OFFSET + self.volume_number * PPM_NAME_LENGTH
        self.volume_name = pascal_to_str(buffer[volume_name_position : volume_name_position + 8])
        self.filename = self.volume_name
        self.prodos_file_type = parent.prodos_file_type
        self.access = parent.access
        self.length = self.blocks_used * BLOCK_SIZE
        self.storage_type = PASCAL_AREA_STORAGE_TYPE
        return self

    @classmethod
    def create(
        cls,
        fs: "ProDOSFilesystem",
        parent: t.Optional["AbstractDirectoryFileEntry"],
        filename: str,
        length: int,  # Length in blocks
        bitmap: "ProDOSBitmap",  # not used
        creation_date: t.Optional[t.Union[date, datetime]] = None,  # not used
        access: int = DEFAULT_ACCESS,  # not used
        file_type: t.Optional[str] = None,  # not used
        aux_type: int = 0,  # not used
        length_bytes: t.Optional[int] = None,  # not used
        resource_length_bytes: t.Optional[int] = None,  # not used
    ) -> "PPMVolumeEntry":
        """
        Create a new Pascal Volume
        """
        # Create the entry
        self = cls(fs, parent)
        self.volume_name = filename
        self.description = ""
        self.default_unit = 0
        self.write_protection = 0
        self.storage_type = PASCAL_AREA_STORAGE_TYPE
        self.blocks_used = length
        self.length = length * BLOCK_SIZE
        self.access = access
        # Write the entry
        if parent is not None:
            parent.update_dir_entry(self, create=True)  # Set volume_number and position
        return self

    def write_buffer(self, buffer: bytearray, position: int = 0) -> None:
        """
        Write the Volume entry to a buffer
        """
        volume_number = position // PPM_INFO_SIZE
        struct.pack_into(
            PPM_INFO_FORMAT,
            buffer,
            position,
            self.key_pointer,
            self.blocks_used,
            self.default_unit,
            self.write_protection,
            0,  # reserved
        )
        # Description
        description_position = PPM_DESCRIPTION_OFFSET + volume_number * PPM_DESCRIPTION_LENGTH
        description = str_to_pascal(self.description)[:PPM_DESCRIPTION_LENGTH]
        buffer[description_position : description_position + len(description)] = description
        # Name
        volume_name_position = PPM_NAME_OFFSET + volume_number * PPM_NAME_LENGTH
        volume_name = str_to_pascal(self.volume_name)[:PPM_NAME_LENGTH]
        buffer[volume_name_position : volume_name_position + len(volume_name)] = volume_name

    def delete(self) -> bool:
        """
        Delete the directory end
        """
        if not isinstance(self.parent, PPMDirectoryEntry):
            return False
        if not self.parent.update_dir_entry(self, delete=True):
            return False
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        if not isinstance(self.parent, PPMDirectoryEntry):
            return False
        if not self.parent.update_dir_entry(self):
            return False
        return True

    def blocks(self, include_indexes: bool = False) -> t.Iterator[int]:
        yield from range(self.key_pointer, self.key_pointer + self.blocks_used)

    def __str__(self) -> str:
        return f"{self.filename:<16} Default unit: #{self.default_unit:<2d}     {self.key_pointer:>7} {self.blocks_used:>7} blocks  Description: {self.description:<16}"

    def __repr__(self) -> str:
        return str(self.__dict__)



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import struct
import typing as t

from ..commons import BLOCK_SIZE
from .block import BlockDevice
from .rx import (
    RX01_SECTOR_SIZE,
    RX02_SECTOR_SIZE,
    get_sector_size,
    rx_extract_12bit_words,
    rx_pack_12bit_words,
    rxfactr_12bit,
)

if t.TYPE_CHECKING:
    from ..abstract import AbstractFile

__all__ = [
    "BlockDevice12Bit",
    "DECtape12Bit",
    "RXBlockDevice12Bit",
    "COSRXBlockDevice12Bit",
]

BYTES_PER_WORD = 2  # Each word is encoded in 2 bytes
BLOCK_SIZE_WORDS = 256  # Size of a block in words (12-bit mode)
COS_TRACKS = 77  # Number of tracks in a COS RX01 disk image
COS_SECTORS_PER_TRACK = 26  # Number of sectors per track in COS RX01 disk image
COS_SECTORS_PER_BLOCK = 3  # Number of sectors per block in COS RX01 disk image
DECTAPE_BLOCK_SIZE = 129  # DECtape block size in words
DECTAPE_BLOCKS_PER_DEVICE_BLOCK = 2  # Number of DECtape blocks per device block


class BlockDevice12Bit(BlockDevice):
    """
    Block device for 12-bit mode
    """

    block_size_words: int = BLOCK_SIZE_WORDS  # Size of a block in bytes
    block_offset_bytes: int = 0  # Offset in bytes for the blocks

    def read_words_block(self, block_number: int) -> t.List[int]:
        """
        Read a block as `block_words` 12bit words
        """
        data = self.read_block(block_number)
        return [x & 0o7777 for x in struct.unpack(f"<{self.block_size_words}H", data)]

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write `block_words` 12bit words as a block
        """
        data = struct.pack(f"<{self.block_size_words}H", *words)
        self.write_block(data, block_number)

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        data = bytearray()
        for i in range(block_number, block_number + number_of_blocks):
            position = i * self.block_size_words * BYTES_PER_WORD + self.block_offset_bytes
            self.f.seek(position)
            data += self.f.read(self.block_size_words * BYTES_PER_WORD)
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        for i in range(block_number, block_number + number_of_blocks):
            position = i * self.block_size_words * BYTES_PER_WORD + self.block_offset_bytes
            self.f.seek(position)
            self.f.write(buffer[: self.block_size_words * BYTES_PER_WORD])
            buffer = buffer[self.block_size_words * BYTES_PER_WORD :]


def cos_sector_position(sector: int, sector_size: int) -> int:
    """
    Calculate the position of a sector in a COS RX01 disk image
    """
    track = 1 + (sector // COS_SECTORS_PER_TRACK)
    sector_in_track = (sector * COS_SECTORS_PER_BLOCK) % COS_SECTORS_PER_TRACK
    return ((track * COS_SECTORS_PER_TRACK) + sector_in_track) * sector_size


class DECtape12Bit(BlockDevice12Bit):
    """
    DECtape block device
    OS/8 only uses 128 words out of 129 in a PDP-8 DECtape block
    """

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        data = bytearray()
        for i in range(number_of_blocks * DECTAPE_BLOCKS_PER_DEVICE_BLOCK):
            self.f.seek(block_number * DECTAPE_BLOCK_SIZE * 4 + i)
            data += self.f.read(BLOCK_SIZE // DECTAPE_BLOCKS_PER_DEVICE_BLOCK)
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        for i in range(number_of_blocks * DECTAPE_BLOCKS_PER_DEVICE_BLOCK):
            self.f.seek(block_number * DECTAPE_BLOCK_SIZE * 4 + i)
            pos = i * BLOCK_SIZE // DECTAPE_BLOCKS_PER_DEVICE_BLOCK
            self.f.write(buffer[pos : pos + BLOCK_SIZE // DECTAPE_BLOCKS_PER_DEVICE_BLOCK])


class RXBlockDevice12Bit(BlockDevice12Bit):
    """
    Block device for 12-bit
    """

    def __init__(self, file: "AbstractFile", cos_support: bool = False):
        super().__init__(file)
        self.sector_size = get_sector_size(self.size)
        self.cos_support = cos_support
        self.is_rx = self.sector_size == RX01_SECTOR_SIZE

    def read_sector(self, sector: int) -> bytes:
        """
        Read a sector from the disk image (for COS RX01 floppy)
        """
        position = cos_sector_position(sector, self.sector_size)
        self.f.seek(position)
        return self.f.read(self.sector_size)

    def write_sector(self, buffer: bytes, sector: int) -> None:
        """
        Write a sector to the disk image (for COS RX01 floppy)
        """
        position = cos_sector_position(sector, self.sector_size)
        self.f.seek(position)
        self.f.write(buffer)

    def read_words_block(self, block_number: int) -> t.List[int]:
        """
        Read a block as `block_size_words` 12bit words
        """
        if self.is_rx:
            # RX01/RX02 12-bit mode
            result = []
            for position in rxfactr_12bit(block_number, self.sector_size):
                self.f.seek(position)
                data = self.f.read(self.sector_size)
                result.extend(rx_extract_12bit_words(data, 0, self.sector_size))
            return result
        else:
            # Non-RX mode, read as block_size_words 12-bit words
            data = self.read_block(block_number)
            return [x & 0o7777 for x in struct.unpack(f"<{self.block_size_words}H", data)]

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write `block_size_words` 12bit words as a block
        """
        if self.is_rx:
            # RX01/RX02 12-bit mode
            if self.sector_size == RX01_SECTOR_SIZE:
                words_per_sector = 64
            elif self.sector_size == RX02_SECTOR_SIZE:
                words_per_sector = 128
            for i, position in enumerate(rxfactr_12bit(block_number, self.sector_size)):
                words_position = i * words_per_sector
                sector_data = rx_pack_12bit_words(words, words_position, self.sector_size)
                self.f.seek(position)
                self.f.write(sector_data)
        else:
            # Non-RX mode, write as `block_size_words` 12-bit words
            data = struct.pack(f"<{self.block_size_words}H", *words)
            self.write_block(data, block_number)


class COSRXBlockDevice12Bit(RXBlockDevice12Bit):
    """
    Block device for 12-bit
    Supports RX01 and RX02 devices and RX01 in COS byte mode
    """

    def __init__(self, file: "AbstractFile"):
        super().__init__(file)
        self.sector_size = get_sector_size(self.size)
        self.is_rx = self.sector_size == RX01_SECTOR_SIZE

    def read_words_block(self, block_number: int) -> t.List[int]:
        """
        Read a block as `block_size_words` 12bit words
        """
        if self.is_rx:
            # COS RX01 byte mode
            sector = block_number * COS_SECTORS_PER_BLOCK
            data = self.read_sector(sector) + self.read_sector(sector + 1) + self.read_sector(sector + 2)
            words: t.List[int] = []
            for i in range(self.sector_size):
                words.append(data[self.sector_size + 2 * i] + ((data[i] >> 4) << 8))
                words.append(data[self.sector_size + 2 * i + 1] + ((data[i] & 0x0F) << 8))
            return words
        else:
            # Non-RX mode, read as `block_size_words` 12-bit words
            data = self.read_block(block_number)
            return [x & 0o7777 for x in struct.unpack(f"<{self.block_size_words}H", data)]

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write `block_size_words` 12bit words as a block
        """
        if self.is_rx:
            # COS RX01 byte mode
            sector = block_number * COS_SECTORS_PER_BLOCK
            buffer = bytearray(self.sector_size * COS_SECTORS_PER_BLOCK)
            for i in range(self.sector_size):
                word1 = words[2 * i]
                word2 = words[2 * i + 1]
                # High bits
                buffer[i] = ((word1 >> 8) << 4) | (word2 >> 8 & 0x0F)
                # Low bits
                buffer[self.sector_size + 2 * i] = word1 & 0xFF
                buffer[self.sector_size + 2 * i + 1] = word2 & 0xFF
            # Write the sectors
            self.write_sector(buffer[: self.sector_size], sector)
            self.write_sector(buffer[self.sector_size : 2 * self.sector_size], sector + 1)
            self.write_sector(buffer[2 * self.sector_size :], sector + 2)
        elif self.is_rx:
            # RX01/RX02 12-bit mode
            if self.sector_size == RX01_SECTOR_SIZE:
                words_per_sector = 64
            elif self.sector_size == RX02_SECTOR_SIZE:
                words_per_sector = 128
            for i, position in enumerate(rxfactr_12bit(block_number, self.sector_size)):
                words_position = i * words_per_sector
                sector_data = rx_pack_12bit_words(words, words_position, self.sector_size)
                self.f.seek(position)
                self.f.write(sector_data)
        else:
            # Non-RX mode, write as `block_size_words` 12-bit words
            data = struct.pack(f"<{self.block_size_words}H", *words)
            self.write_block(data, block_number)



from xferx.commons import ASCII, IMAGE, READ_FILE_FULL
from xferx.device.block_18bit import (
    from_18bit_words_to_bytes,
    from_bytes_to_18bit_words,
)
from xferx.pdp15.adssfs import (
    DIRECTORY_BLOCK,
    WORDS_PER_BLOCK,
    ADSSDirectory,
    ADSSDirectoryEntry,
    ADSSFilesystem,
    ascii_to_five_seven,
    decode_block_format,
    encode_block_format,
    five_seven_to_ascii,
    oct_dump,
)
from xferx.shell import Shell

DSK = "tests/dsk/adss.dtp"

# fmt: off
DIR = [
    0o777777, 0o777777, 0o777777, 0o777777, 0o777471, 0o630614, 0o306143, 0o061430,
    0o614306, 0o143041, 0o020410, 0o204102, 0o041021, 0o430614, 0o306143, 0o061430,
    0o614306, 0o143061, 0o430614, 0o306143, 0o061430, 0o614306, 0o143061, 0o437777,
    0o777777, 0o777777, 0o777777, 0o777777, 0o777777, 0o777777, 0o777777, 0o777777,
    0o561417, 0o010400, 0o021116, 0o400037, 0o040424, 0o000000, 0o021116, 0o400040,
    0o053005, 0o032524, 0o021116, 0o400041, 0o111624, 0o050105, 0o021116, 0o400043,
    0o111624, 0o161716, 0o021116, 0o400050, 0o220514, 0o050105, 0o021116, 0o400076,
    0o220514, 0o161716, 0o021116, 0o400077, 0o561411, 0o022200, 0o021116, 0o400105,
    0o061703, 0o011400, 0o021116, 0o400120, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o777777, 0o777742, 0o104777, 0o777777, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o007777,
    0o777777, 0o777777, 0o777777, 0o777777, 0o777777, 0o777777, 0o777777, 0o777777,
    0o131571, 0o556165, 0o233123, 0o400000, 0o231320, 0o021413, 0o233123, 0o400042,
    0o111702, 0o141300, 0o233123, 0o400046, 0o230716, 0o021413, 0o233123, 0o400052,
    0o233123, 0o100116, 0o233123, 0o400056, 0o233123, 0o021413, 0o233123, 0o400061,
    0o562331, 0o231404, 0o233123, 0o400062, 0o021124, 0o150120, 0o233123, 0o400071,
    0o041122, 0o050324, 0o233123, 0o400100, 0o050411, 0o240000, 0o233123, 0o400644,
    0o201120, 0o000000, 0o233123, 0o400664, 0o150103, 0o221700, 0o233123, 0o400704,
    0o066400, 0o000000, 0o233123, 0o400742, 0o042403, 0o172031, 0o233123, 0o401001,
    0o042515, 0o200000, 0o233123, 0o401004, 0o252004, 0o012405, 0o233123, 0o401010,
    0o230705, 0o160000, 0o233123, 0o401020, 0o031001, 0o111600, 0o233123, 0o401051,
    0o200124, 0o031000, 0o233123, 0o401071, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
    0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000, 0o000000,
]
# fmt: on


class MockFilesytem(ADSSFilesystem):

    def __init__(self, t):
        self.t = t

    def read_words_block(self, block):
        block_number = block if isinstance(block, int) else block.block_number
        if block_number in self.t:
            return self.t[block_number]
        else:
            raise ValueError(f"Invalid block number {block_number}")

    def write_words_block(self, block, words) -> None:
        block_number = block if isinstance(block, int) else block.block_number
        self.t[block_number] = words


def test_five_seven():
    """
    Test conversion between ASCII and 5-7 encoding
    """
    data = [f"{i:>5} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890" for i in range(10)]
    for line in data:
        lineb = line.encode("ascii")
        words = ascii_to_five_seven(lineb)
        tmp = five_seven_to_ascii(words)
        assert lineb == tmp.rstrip(b"\0")


def test_18bit_words():
    words = list(range(0, 2048))
    tmp = from_18bit_words_to_bytes(words, IMAGE)
    w2 = from_bytes_to_18bit_words(tmp, IMAGE)

    words = list(range(0, 127)) * 26
    tmp = from_18bit_words_to_bytes(words, ASCII)
    w2 = from_bytes_to_18bit_words(tmp, ASCII)
    assert words == w2


def encode_decode(content, mode, l=None):
    blocks_content = list(encode_block_format(content, mode, words_per_block=WORDS_PER_BLOCK - 1))
    if l is not None:
        assert len(blocks_content) == l
    tmp = bytearray()
    for block_content in blocks_content:
        tmp.extend(decode_block_format(block_content))
    assert bytes(tmp) == content


def test_block_format():
    """
    Test encoding and decoding of block format
    """

    content = b"***Test***\n"
    encode_decode(content, ASCII, 1)

    content = ("\n".join([f"{i:>5} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890" for i in range(50)]) + "\n").encode("ascii")
    encode_decode(content, ASCII, 5)

    content = b'\xb3\x80\xb8\xb5\xa3\xa3\xa6\x80\x96\xa6\xa9\xa3\xb4\x80\xa7\xa9\xa6\xb7\xa9\xb1\xa4\x8c\x80\x81\x9e\x96\xa9\xb9\x93\xb5\x80\x82\x9b\x80\x81\x90\x8c\x80\x82\x81\x90\x9e\xb6\xa6\xa9\xa4\xb1\x93\x80\xaf\x81\x82\xb8\x80\xb8\xb5\xa3\xa3\xa6\x9b\x80\x96\xa6\xa9\xa3\xb4\xad\x8f\x8f\x8c\x80\x83\x9e\xb5\xa5\xb4\x8f\x8f\x8c\x80\x84\x8d\x80\x81\x8e\x80\x80\x80\x80\x80'
    words = list(range(0, 10))
    content = from_18bit_words_to_bytes(words, IMAGE)
    encode_decode(content, IMAGE, 1)

    words = list(range(0, 127))
    content = from_18bit_words_to_bytes(words, IMAGE)
    encode_decode(content, IMAGE, 1)

    words = list(range(0, 1000))
    content = from_18bit_words_to_bytes(words, IMAGE)
    encode_decode(content, IMAGE, 5)


def test_directory():
    fs = MockFilesytem(
        {
            DIRECTORY_BLOCK: list(DIR),
        }
    )
    d = ADSSDirectory.read(fs)

    words = [0o777777, 0o777777, 0o777777, 0o777777]
    e1 = ADSSDirectoryEntry.read(d, words, 0, 0)
    oct_dump(words)
    oct_dump(e1.to_words())
    assert e1.to_words() == words

    d.write()
    print("-----")
    oct_dump(DIR)
    print("-----")
    oct_dump(fs.read_words_block(DIRECTORY_BLOCK))
    assert fs.read_words_block(DIRECTORY_BLOCK) == list(DIR)


def test_adss():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount adss1: /adss {DSK}", batch=True)
    fs = shell.volumes.get('ADSS1')
    assert isinstance(fs, ADSSFilesystem)

    shell.onecmd("dir adss1:", batch=True)
    shell.onecmd("type adss1:f1;src", batch=True)

    x = fs.read_bytes("b100;bin")
    assert len(x) == 300
    assert from_bytes_to_18bit_words(x, IMAGE) == list(range(0, 100))

    x = fs.read_bytes("f100;src")
    assert len(x) == 4400
    for i in range(0, 100):
        assert f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890".encode("ascii") in x

    l = list(fs.filter_entries_list("*;SRC"))
    assert len(l) == 3


def test_adss_init():
    shell = Shell(verbose=True)
    shell.onecmd(f"mount adss1: /adss {DSK}", batch=True)
    shell.onecmd(f"create /allocate:1152 {DSK}.mo", batch=True)
    shell.onecmd(f"init /adss {DSK}.mo", batch=True)
    shell.onecmd(f"mount adss2: /adss {DSK}.mo", batch=True)
    shell.onecmd("dir adss2:", batch=True)
    shell.onecmd("copy adss1:*;SRC adss2:", batch=True)
    fs = shell.volumes.get('ADSS2')

    l = list(fs.filter_entries_list("*;SRC"))
    assert len(l) == 3

    shell.onecmd("delete adss2:F1;SRC", batch=True)
    l = list(fs.filter_entries_list("*;SRC"))
    assert len(l) == 2


def test_adss_write_file():
    shell = Shell(verbose=True)
    shell.onecmd(f"copy {DSK} {DSK}_2.mo", batch=True)
    shell.onecmd(f"mount adss2: /adss {DSK}_2.mo", batch=True)
    fs = shell.volumes.get('ADSS2')
    assert isinstance(fs, ADSSFilesystem)

    for l in (1, 10, 100, 1000):
        words = list(range(0, l))
        data = from_18bit_words_to_bytes(words, IMAGE)
        filename = f"T{l};BIN"
        fs.write_bytes(filename, data)
        data_read = fs.read_bytes(filename)
        assert data_read == data

    for l in (1, 10, 100, 1000):
        filename = f"T{l};BIN"
        data = fs.read_bytes(filename)
        words = from_bytes_to_18bit_words(data, IMAGE)
        assert words == list(range(0, l))

    for l in (1, 10, 100, 1000):
        tmp = bytearray()
        for i in range(0, l):
            tmp.extend(f"{i:5d} ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890\n".encode("ascii"))
        data = bytes(tmp)
        filename = f"T{l};SRC"
        fs.write_bytes(filename, data)
        data_read = fs.read_bytes(filename)
        assert data_read == data

    shell.onecmd("create /allocate:400 adss2:big;bin", batch=True)
    big = fs.get_file_entry("BIG;BIN")
    assert len(list(big.get_blocks())) == 400

    shell.onecmd("dismount adss2:", batch=True)
    shell.onecmd(f"del {DSK}_2.mo", batch=True)



# Copyright (C) 2414 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import math
import os
import struct
import sys
import typing as t
from abc import ABC, abstractmethod
from datetime import date, datetime, timedelta
from functools import reduce

from ..abstract import AbstractBlockFilesystem, AbstractDirectoryEntry, AbstractFile
from ..commons import (
    BLOCK_SIZE,
    READ_FILE_FULL,
    dump_struct,
    filename_match,
    swap_words,
)
from ..device.abstract import AbstractDevice

__all__ = [
    "UNIXFile",
    "UNIXDirectory",
    "UNIXDirectoryEntry",
    "UNIXFilesystem",
]

# def unix_canonical_filename(fullname: str, wildcard: bool = False) -> str:
#     """
#     Generate the canonical unix name
#     """
#     # TODO
#     if fullname:
#         fullname = fullname[:V4_FILENAME_LEN]
#     return fullname


def unix_join(a: str, *p: str) -> str:
    """
    Join two or more pathname components
    """
    path = a
    if not p:
        path[:0] + "/"
    for b in p:
        if b.startswith("/"):
            path = b
        elif not path or path.endswith("/"):
            path += b
        else:
            path += "/" + b
    return path


def unix_split(p: str) -> t.Tuple[str, str]:
    """
    Split a pathname
    """
    i = p.rfind("/") + 1
    head, tail = p[:i], p[i:]
    if head and head != "/" * len(head):
        head = head.rstrip("/")
    return head, tail


def l3tol(data: bytes, n: int) -> t.List[int]:
    """
    Convert 3-byte integers
    """
    result: t.List[int] = []
    for i in range(0, n * 3, 3):
        tmp = (data[i + 1] << 0) + (data[i + 2] << 8) + (data[i + 0] << 16)
        result.append(tmp)
    return result


def iterate_words(data: bytes) -> t.Iterator[int]:
    """
    Iterate over words in a byte array
    """
    for i in range(0, len(data), 2):
        yield struct.unpack("H", data[i : i + 2])[0]


def iterate_long(data: bytes) -> t.Iterator[int]:
    """
    Iterate over longs in a byte array
    """
    for i in range(0, len(data), 4):
        yield swap_words(struct.unpack("I", data[i : i + 4])[0])


def format_mode(flags: int, version: int, perms: t.List[t.List[t.Tuple[int, str]]]) -> str:
    result = []

    for column in perms:
        ch = [ch for flag, ch in column if (flags & flag) == flag]
        if ch:
            result.append(ch[0])
    return "".join(result)


def format_time(t: int) -> str:
    mod_time = datetime.fromtimestamp(t)
    six_months_ago = datetime.now() - timedelta(days=6 * 30)
    if mod_time > six_months_ago:
        return mod_time.strftime("%b %d %H:%M")
    else:
        return mod_time.strftime("%b %d %Y ")


class UNIXFile(AbstractFile):
    inode: "UNIXInode"
    closed: bool

    def __init__(self, inode: "UNIXInode"):
        self.inode = inode
        self.closed = False

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.inode.get_length()
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.inode.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        # Get the blocks to be read
        blocks = list(self.inode.blocks())[block_number : block_number + number_of_blocks]
        # Read the blocks
        for disk_block_number in blocks:
            tmp = self.inode.fs.read_block(disk_block_number)
            data.extend(tmp)
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.inode.get_length()
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        # Get the blocks to be written
        blocks = list(self.inode.blocks())[block_number : block_number + number_of_blocks]
        # Write the blocks
        for i, disk_block_number in enumerate(blocks):
            data = buffer[i * BLOCK_SIZE : (i + 1) * BLOCK_SIZE]
            self.inode.fs.write_block(data, disk_block_number)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.inode.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.inode.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return str(self.inode)


class UNIXInode(ABC):

    fs: "UNIXFilesystem"
    inode_num: int  #      inode number
    flags: int  #          flags
    nlinks: int  #         number of links to file
    uid: int  #            user ID of owner
    gid: t.Optional[int]  # group ID of owner
    size: int  #           size
    addr: t.List[int]  #     block numbers or device numbers
    atime: int = 0  #      time of last access
    mtime: int = 0  #      time of last modification
    ctime: int = 0  #      time of last change to the inode

    def __init__(self, fs: "UNIXFilesystem"):
        self.fs = fs

    @classmethod
    @abstractmethod
    def read(cls, fs: "UNIXFilesystem", inode_num: int, buffer: bytes, position: int = 0) -> "UNIXInode":
        pass

    def write(self) -> None:
        """
        Write inode
        """
        self.fs.dev.f.seek(BLOCK_SIZE * 2 + (self.inode_num - 1) * self.fs.inode_size)
        buffer = bytearray(self.fs.inode_size)
        self.write_buffer(buffer)
        self.fs.dev.f.write(buffer)
        self.fs.dev.f.flush()

    def write_buffer(self, buffer: bytearray, position: int = 0) -> None:
        """
        Write the inode to the buffer
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    @abstractmethod
    def blocks(self) -> t.Iterator[int]:
        pass

    @property
    @abstractmethod
    def isdir(self) -> bool:
        pass

    @property
    @abstractmethod
    def is_regular_file(self) -> bool:
        pass

    @property
    @abstractmethod
    def is_allocated(self) -> bool:
        pass

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return int(math.ceil(self.get_size() / self.get_block_size()))

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.size

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return BLOCK_SIZE

    def __repr__(self) -> str:
        return str(self.__dict__)


class UNIXDirectoryEntry(AbstractDirectoryEntry):

    fs: "UNIXFilesystem"
    _inode: t.Optional["UNIXInode"]
    inode_num: int  # Inode number
    filename: str  # File name
    dirname: str  # Parent directory name

    def __init__(self, fs: "UNIXFilesystem", fullname: str, inode_num: int, inode: t.Optional["UNIXInode"] = None):
        self.fs = fs
        self.dirname, self.filename = unix_split(fullname)
        self.inode_num = inode_num
        self._inode = inode

    @property
    def inode(self) -> "UNIXInode":
        if self._inode is None:
            self._inode = self.fs.read_inode(self.inode_num)
        return self._inode

    @property
    def is_empty(self) -> bool:
        return not self.inode.is_allocated

    @property
    def fullname(self) -> str:
        return unix_join(self.dirname, self.filename)

    @property
    def basename(self) -> str:
        return self.filename

    def get_length(self) -> int:
        """Get the length in blocks"""
        return self.inode.get_length()

    def get_size(self) -> int:
        """Get file size in bytes"""
        return self.inode.get_size()

    def get_block_size(self) -> int:
        """Get file block size in bytes"""
        return self.inode.get_block_size()

    @property
    def creation_date(self) -> t.Optional[date]:
        return datetime.fromtimestamp(self.inode.mtime)

    def delete(self) -> bool:
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def write(self) -> bool:
        """
        Write the directory entry
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def open(self, file_mode: t.Optional[str] = None) -> UNIXFile:
        """
        Open a file
        """
        return UNIXFile(self.inode)

    def __lt__(self, other: "UNIXDirectoryEntry") -> bool:
        return self.filename < other.filename

    def __gt__(self, other: "UNIXDirectoryEntry") -> bool:
        return self.filename > other.filename

    def __str__(self) -> str:
        return f"{self.inode_num:>5} {self.basename}"


class UNIXDirectory:

    fs: "UNIXFilesystem"
    inode: "UNIXInode"
    entries: t.List[t.Tuple[int, str]]

    def __init__(self, fs: "UNIXFilesystem", inode: "UNIXInode"):
        self.fs = fs
        self.inode = inode

    @classmethod
    def read(cls, fs: "UNIXFilesystem", inode: "UNIXInode") -> "UNIXDirectory":
        """
        Read the directory
        """
        self = UNIXDirectory(fs, inode)
        if self.inode.isdir:
            self.entries = []
            f = UNIXFile(self.inode)
            try:
                while True:
                    data = f.read(struct.calcsize(self.fs.dir_format))
                    inode_num, name = struct.unpack_from(self.fs.dir_format, data)
                    name_ascii = name.decode("ascii", errors="ignore").rstrip("\x00")
                    self.entries.append((inode_num, name_ascii))
            except IOError:
                pass
            finally:
                f.close()
        return self

    def write(self) -> None:
        """
        Write the directory
        """
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))


class UNIXFilesystem(AbstractBlockFilesystem):
    """
    UNIX Filesystem
    """

    version: int  # UNIX version
    inode_size: int  #
    dir_format: str
    root_inode: int  # Root inode number
    unix_inode_class: t.Type["UNIXInode"]
    directory_class: t.Type["UNIXDirectory"] = UNIXDirectory
    pwd: str
    inode_list_blocks: int  # number of blocks devoted to the i-list
    volume_size: int  # size in blocks of entire volume
    free_blocks_in_list: int  # number of free blocks in the free list
    free_inodes_in_list: int  # number of free i-numbers in the inode array
    inodes: int = 0  # number of inodes
    first_inode: int = 1  # first inode number
    perms: t.List[t.List[t.Tuple[int, str]]]  # permissions

    @classmethod
    @abstractmethod
    def mount(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "UNIXFilesystem":
        """
        Mount the filesystem from a file or device
        """
        pass

    def read_inode(self, inode_num: int) -> UNIXInode:
        """
        Read inode by number
        """
        self.dev.f.seek(BLOCK_SIZE * 2 + (inode_num - 1) * self.inode_size)
        data = self.dev.f.read(self.inode_size)
        return self.unix_inode_class.read(self, inode_num, data)

    def get_inode(self, path: str) -> t.Optional["UNIXInode"]:
        """
        Get inode by path
        """
        path = unix_join(self.pwd, path) if not path.startswith("/") else path
        parts = path.split("/")
        inode_num = self.root_inode

        while True:
            # Read inode
            inode = self.read_inode(inode_num)
            if not parts:
                # No more parts, inode found
                return inode if inode.is_allocated else None
            elif not inode.isdir:
                # More parts and not a directory, not found
                return None
            # Get next part
            name = parts.pop(0)
            if name:
                # Search for the name in the directory
                found = False
                for no, nm in self.list_dir(inode):
                    if no > 0 and nm == name:
                        inode_num = no
                        found = True
                        break
                if not found:
                    return None

    def list_dir(self, inode: UNIXInode) -> t.List[t.Tuple[int, str]]:
        directory = self.directory_class.read(self, inode)
        return directory.entries

    def read_dir_entries(self, dirname: str) -> t.Iterator["UNIXDirectoryEntry"]:
        inode = self.get_inode(dirname)
        if inode:
            for inode_num, filename in self.list_dir(inode):
                if inode_num > 0:
                    fullname = unix_join(dirname, filename)
                    yield UNIXDirectoryEntry(self, fullname, inode_num)

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["UNIXDirectoryEntry"]:
        if not pattern and expand:
            pattern = "*"
        if pattern and pattern.startswith("/"):
            absolute_path = pattern
        else:
            absolute_path = unix_join(self.pwd, pattern or "")
        if self.isdir(absolute_path):
            if not expand:
                yield self.get_file_entry(absolute_path)  # type: ignore
                return
            dirname = pattern
            pattern = "*"
        else:
            dirname, pattern = unix_split(absolute_path)
        for entry in self.read_dir_entries(dirname):  # type: ignore
            if filename_match(entry.basename, pattern, wildcard):
                yield entry

    @property
    def entries_list(self) -> t.Iterator["UNIXDirectoryEntry"]:
        yield from self.read_dir_entries(self.pwd)

    def get_file_entry(self, fullname: str) -> UNIXDirectoryEntry:
        inode = self.get_inode(fullname)
        if not inode:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)
        return UNIXDirectoryEntry(self, fullname, inode.inode_num, inode)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[UNIXDirectoryEntry]:
        raise OSError(errno.EROFS, os.strerror(errno.EROFS))

    def isdir(self, fullname: str) -> bool:
        inode = self.get_inode(fullname)
        return inode is not None and inode.isdir

    def read_uids(self) -> t.Dict[int, str]:
        """
        Read the uid -> name map
        """
        result: t.Dict[int, str] = {}
        filename = "/etc/uids" if self.version < 3 else "/etc/passwd"
        try:
            for line in self.read_text(filename).split("\n"):
                if self.version < 3:
                    try:
                        name, uid = line.split(":", 1)
                        result[int(uid)] = name
                    except Exception:
                        pass
                else:
                    try:
                        name, _, uid, _ = line.split(":", 3)
                        result[int(uid)] = name
                    except Exception:
                        pass
        except Exception:
            pass
        return result

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        entries = sorted(self.filter_entries_list(pattern, include_all=True, wildcard=True))
        if not entries:
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), pattern)
        uids = self.read_uids()
        if not options.get("brief"):
            blocks = reduce(lambda x, y: x + y, [x.inode.get_length() for x in entries])
            if self.version < 3:
                sys.stdout.write(f"total {blocks:>4}\n")
            else:
                sys.stdout.write(f"blocks = {blocks}\n")
        for x in entries:
            if not options.get("full") and x.basename.startswith("."):
                pass
            elif options.get("brief"):
                # Lists only file names
                sys.stdout.write(f"{x.basename}\n")
            else:
                mode = format_mode(x.inode.flags, self.version, self.perms)
                time = format_time(x.inode.mtime)
                uid = uids.get(x.inode.uid, str(x.inode.uid))
                if self.version < 3:
                    sys.stdout.write(
                        f"{x.inode_num:>3} {mode} {x.inode.nlinks:>2} {uid:<6} {x.inode.size:>6} {time} {x.basename}\n"
                    )
                else:
                    sys.stdout.write(
                        f"{x.inode_num:>5} {mode}{x.inode.nlinks:>2} {uid:<6}{x.inode.size:>7} {time} {x.basename}\n"
                    )

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if arg:
            if arg.isnumeric():
                # Dump the inode by number
                inode_num = int(arg)
                inode: t.Optional[UNIXInode] = self.read_inode(inode_num)
            else:
                # Dump the inode by path
                inode = self.get_inode(arg)
            if inode:
                if hasattr(inode, "examine"):
                    sys.stdout.write(inode.examine())  # type: ignore
                else:
                    sys.stdout.write(dump_struct(inode.__dict__) + "\n")
                if inode.isdir:
                    # Dump the directory entries
                    sys.stdout.write("Directory entries:\n")
                    for inode_num, filename in self.list_dir(inode):
                        if inode_num > 0:
                            child_inode: t.Optional[UNIXInode] = self.read_inode(inode_num)
                            sys.stdout.write(f"{child_inode} {filename}\n")
        else:
            # Dump the entire filesystem
            sys.stdout.write(dump_struct(self.__dict__))
            sys.stdout.write("\n")
            for i in range(1, self.inodes + 1):
                inode = self.read_inode(i)
                sys.stdout.write(f"{inode}\n")

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def chdir(self, fullname: str) -> bool:
        """
        Change the current directory
        """
        if not fullname.startswith("/"):
            fullname = unix_join(self.pwd, fullname)
        fullname = os.path.normpath(fullname)
        if self.isdir(fullname):
            self.pwd = fullname
            return True
        else:
            return False

    def get_pwd(self) -> str:
        """
        Get the current directory
        """
        return self.pwd


class Bitmap:

    # fs: "ProDOSFilesystem"
    bitmaps: t.List[int]
    # bitmap_blocks: int  # Number of bitmap blocks

    # def __init__(self, fs: "ProDOSFilesystem"):
    #     self.fs = fs
    #
    # @classmethod
    # def read(cls, fs: "ProDOSFilesystem") -> "ProDOSBitmap":
    #     """
    #     Read the bitmap blocks
    #     """
    #     self = ProDOSBitmap(fs)
    #     self.bitmap_blocks = cls.calculate_bitmap_size(fs)
    #     # Read the bitmap blocks
    #     self.bitmaps = []
    #     for block_number in range(fs.bit_map_pointer, fs.bit_map_pointer + self.bitmap_blocks):
    #         buffer = fs.read_block(block_number)
    #         if not buffer:
    #             raise OSError(errno.EIO, os.strerror(errno.EIO))
    #         self.bitmaps += list(buffer)
    #     return self
    #
    # @classmethod
    # def create(cls, fs: "ProDOSFilesystem") -> "ProDOSBitmap":
    #     """
    #     Create the bitmap blocks
    #     """
    #     self = ProDOSBitmap(fs)
    #     self.bitmap_blocks = cls.calculate_bitmap_size(fs)
    #     self.bitmaps = [0] * BLOCK_SIZE * self.bitmap_blocks
    #     # Mark the bitmap blocks of the volume as free
    #     for i in range(0, int(math.ceil(self.fs.get_size() / BLOCK_SIZE))):
    #         self.set_free(i)
    #     # Mark the first blocks as used
    #     for i in range(0, VOLUME_DIRECTORY_BLOCK + DEFAULT_DIRECTORY_BLOCKS + 1):
    #         self.set_used(i)
    #     return self

    # @classmethod
    # def calculate_bitmap_size(cls, fs: "ProDOSFilesystem") -> int:
    #     """
    #     Calculate the number of blocks in the bitmap
    #     """
    #     bitmap_bytes = fs.total_blocks // 8
    #     if fs.total_blocks % 8 > 0:
    #         bitmap_bytes += 1
    #     bitmap_blocks = bitmap_bytes // BLOCK_SIZE
    #     if bitmap_bytes % BLOCK_SIZE > 0:
    #         bitmap_blocks += 1
    #     return bitmap_blocks

    # def write(self) -> None:
    #     """
    #     Write the bitmap blocks
    #     """
    #     for i in range(0, self.bitmap_blocks):
    #         buffer = bytes(self.bitmaps[i * BLOCK_SIZE : (i + 1) * BLOCK_SIZE])
    #         self.fs.write_block(buffer, self.fs.bit_map_pointer + i)

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return len(self.bitmaps) * 8

    def is_free(self, block_number: int) -> bool:
        """
        Check if the block is free
        """
        int_index = block_number // 8
        bit_position = 7 - (block_number % 8)
        bit_value = self.bitmaps[int_index]
        return (bit_value & (1 << bit_position)) != 0

    def set_free(self, block_number: int) -> None:
        """
        Mark the block as free
        """
        int_index = block_number // 8
        bit_position = 7 - (block_number % 8)
        self.bitmaps[int_index] |= 1 << bit_position

    def set_used(self, block_number: int) -> None:
        """
        Mark the block as used
        """
        int_index = block_number // 8
        bit_position = 7 - (block_number % 8)
        self.bitmaps[int_index] &= ~(1 << bit_position)

    def allocate(self, size: int) -> t.List[int]:
        """
        Allocate blocks
        """
        blocks = []
        for block in range(0, self.total_bits):
            if self.is_free(block):
                self.set_used(block)
                blocks.append(block)
            if len(blocks) == size:
                break
        if len(blocks) < size:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))
        return blocks

    def used(self) -> int:
        """
        Count the number of used blocks
        """
        return self.total_bits - self.free()

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        free = 0
        for block in self.bitmaps:
            free += block.bit_count()
        return free


class InodeBitmap:

    bitmaps: t.List[int]

    @property
    def total_bits(self) -> int:
        """
        Return the bitmap length in bit
        """
        return len(self.bitmaps) * 8

    def is_free(self, block_number: int) -> bool:
        """
        Check if the inode is free
        """
        int_index = block_number // 8
        bit_position = 7 - (block_number % 8)
        bit_value = self.bitmaps[int_index]
        return (bit_value & (1 << bit_position)) == 0

    def set_free(self, block_number: int) -> None:
        """
        Mark the inode as free
        """
        int_index = block_number // 8
        bit_position = 7 - (block_number % 8)
        self.bitmaps[int_index] &= ~(1 << bit_position)

    def set_used(self, block_number: int) -> None:
        """
        Mark the inode as used
        """
        int_index = block_number // 8
        bit_position = 7 - (block_number % 8)
        self.bitmaps[int_index] |= 1 << bit_position

    def allocate(self) -> int:
        """
        Allocate an inode
        """
        for block in range(0, self.total_bits):
            if self.is_free(block):
                self.set_used(block)
                return block
        raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC))

    def free(self) -> int:
        """
        Count the number of free inodes
        """
        return self.total_bits - self.used()

    def used(self) -> int:
        """
        Count the number of used inodes
        """
        free = 0
        for block in self.bitmaps:
            free += block.bit_count()
        return free



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import string
import traceback
import typing as t

from ..abstract import AbstractDirectoryEntry, AbstractFilesystem

if t.TYPE_CHECKING:
    from ..shell import Shell

__all__ = [
    "CommandError",
    "Commands",
    "PartialMatching",
    "ShellCommand",
    "add_slash",
    "ask",
    "copy_file",
    "extract_options",
    "get_int_option",
    "split_command_line",
]

IDENTCHARS = string.ascii_letters + string.digits + "@_"


class CommandError(Exception):
    """
    Exception raised for command errors
    """

    def __init__(self, message: str) -> None:
        super().__init__(message)
        self.message = message

    def __str__(self) -> str:
        return self.message


def ask(prompt: str) -> str:
    """
    Prompt the user for input with the given prompt message
    """
    result = ""
    while not result:
        result = input(prompt).strip()
    return result


def extract_options(args: t.List[str], *options: str) -> t.Tuple[t.List[str], t.Dict[str, t.Union[bool, str]]]:
    """
    Extract options from the command line
    """
    result: t.List[str] = []
    options_result: t.Dict[str, t.Union[bool, str]] = {}
    for arg in args:
        if ':' in arg:
            key, value = arg.split(':', 1)
            if key.lower() in options:
                options_result[key.lower()[1:]] = value
            else:
                result.append(arg)
        elif arg.lower() in options:
            options_result[arg.lower()[1:]] = True
        else:
            result.append(arg)
    return result, options_result


def get_int_option(
    options: t.Dict[str, t.Union[bool, str]], key: str, default: t.Optional[int] = None
) -> t.Optional[int]:
    """
    Get an integer option from the options dictionary
    """
    try:
        value = int(options[key])
        if value < 0:
            raise ValueError
        return value
    except KeyError:
        return default
    except ValueError:
        raise CommandError("?KMON-F-Invalid value specified with option")


def copy_file(
    from_fs: AbstractFilesystem,
    from_entry: AbstractDirectoryEntry,
    to_fs: AbstractFilesystem,
    to_path: str,
    file_type: t.Optional[str],
    file_mode: t.Optional[str],
    verbose: int,
    cmd: str = "COPY",
) -> None:
    if not file_type:
        file_type = from_entry.file_type
    try:
        content = from_entry.read_bytes(file_mode)
        to_fs.write_bytes(to_path, content, from_entry.creation_date, file_type, file_mode)
    except Exception:
        if verbose:
            traceback.print_exc()
        raise CommandError(f"?{cmd}-F-Error copying {from_entry.fullname}")


def add_slash(fs: AbstractFilesystem, filename: str) -> str:
    """
    Add a trailing slash to a filename if it is a directory
    """
    try:
        if fs.isdir(filename):
            filename = filename + "/"
        return filename.replace(" ", "\\ ")
    except Exception:
        pass
    return filename


def split_command_line(line: str) -> t.Tuple[t.Optional[str], t.Optional[str]]:
    """
    Split a command line into command and argument
    """
    line = line.strip()
    if not line or line.startswith("#"):
        return None, None
    elif line[0] == '?':
        line = f"help {line[1:]}"
    elif line[0] == '!':
        line = f"shell {line[1:]}"
    elif line[0] == '@':
        line = f"batch {line[1:]}"
    i, n = 0, len(line)
    while i < n and line[i] in IDENTCHARS:
        i = i + 1
    cmd, arg = line[:i], line[i:].strip()
    return cmd.upper(), arg


class PartialMatching:
    """
    A utility class for managing keys that can be partially matched based on prefixes.
    """

    def __init__(self) -> None:
        self.short: t.Dict[str, str] = {}  # short key => full key
        self.full: t.Dict[str, str] = {}  # full key => short key

    def add(self, key: str) -> None:
        """
        Add a key to the partial matching dictionary
        A key is constituted by a prefix and a tail, separated by an underscore.
        Example: DIR_ECTORY
        """
        key = key.upper()
        try:
            prefix, tail = key.split("_", 1)
        except:
            prefix = key
            tail = ""
        full = prefix + tail
        self.full[full] = prefix
        self.short[prefix] = full

    def add_alias(self, full: str, alias: str) -> None:
        """
        Add an alias
        """
        self.short[alias.upper()] = full.upper()

    def get(self, key: str, default: t.Optional[str] = None) -> t.Optional[str]:
        """
        Retrieves the full key corresponding to a given prefix or alias.
        """
        key = key.upper()
        try:
            return self.short[key]
        except KeyError:
            pass
        matching_keys = [(k, v) for k, v in self.full.items() if k.startswith(key) and len(key) >= len(v)]
        if not matching_keys:
            return default
        return matching_keys[0][0]


ShellCommand = t.Callable[["Shell", t.List[str]], None]


class Commands:

    commands: t.Dict[str, ShellCommand] = {}
    cmd_matching: PartialMatching

    def __init__(self) -> None:
        self.commands = {}
        self.cmd_matching = PartialMatching()

    def register(self, decorator_arg: str, aliases: t.List[str] = []) -> t.Callable[[ShellCommand], ShellCommand]:
        """
        Register a command function with the shell
        """

        def decorator(func: ShellCommand) -> ShellCommand:
            if " " not in decorator_arg:
                # Commands without spaces are considered main commands
                if "_" in decorator_arg:
                    self.cmd_matching.add(decorator_arg)
                command = decorator_arg.replace("_", "")
                self.commands[command] = func
                for alias in aliases:
                    self.cmd_matching.add_alias(command, alias)
            else:
                # Subcommands have a space in the decorator argument
                main, sub = decorator_arg.split(" ", 1)
                if "_" in decorator_arg:
                    self.cmd_matching.add(decorator_arg)
                command = decorator_arg.replace("_", "")
                self.commands[command] = func
                for alias in aliases:
                    self.cmd_matching.add_alias(command, alias)
            return func

        return decorator

    def get(self, key: str) -> t.Optional[ShellCommand]:
        """
        Get a command function by its complete/abbreviated name
        """
        key = key.upper()
        return self.commands.get(self.cmd_matching.get(key) or key)

    def get_commands(self, exclude: t.List[str] = []) -> t.List[str]:
        """
        Get a list of all registered commands, excluding those in the exclude list
        """
        return [k for k in self.commands if k not in exclude]



# Copyright (C) 2014 Andrea Bonomi <andrea.bonomi@gmail.com>

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

import errno
import io
import math
import os
import re
import sys
import typing as t
from datetime import date

from ..abstract import AbstractDirectoryEntry, AbstractFile, AbstractFilesystem
from ..commons import ASCII, BLOCK_SIZE, DECTAPE, IMAGE, READ_FILE_FULL, filename_match
from ..device.abstract import AbstractDevice
from ..device.block_12bit import BlockDevice12Bit, DECtape12Bit, RXBlockDevice12Bit
from ..device.rx import RX_SECTOR_TRACK

__all__ = [
    "OS8File",
    "OS8DirectoryEntry",
    "OS8Filesystem",
]

# OS/8 Software Support Manual
# http://www.bitsavers.org/pdf/dec/pdp8/os8/DEC-S8-OSSMB-A-D_OS8_v3ssup.pdf Pag 62

DIR_ENTRY_SIZE = 5  # Directory entry size (words)
EMPTY_DIR_ENTRY_SIZE = 2  # Empty directory entry size (words)
DIRECTORY_SEGMENT_HEADER_SIZE = 5  # Directory segment header (words)
DIRECTORY_SEGMENT_SIZE = 256  # Directory segment size (words)
DIRECTORY_SEGMENT_START = 1  # Directory start block number
NUM_OF_SEGMENTS = 6  # Number of directory segments
PARTITION_FULLNAME_RE = re.compile(r"^\[(\d+)\](.*)$")
ASCII_EXTENSIONS = [
    "BA",  # BASIC
    "BI",  # BATCH
    "FC",  # FOCAL
    "FT",  # FORTRAN
    "HL",  # HELP
    "LS",  # Listing
    "MA",  # MACRO
    "PA",  # PAL
    "PS",  # Pascal
    "RA",  # RALF
    "SB",  # SABR
    "TE",  # TECO
    "TX",  # Text
    "WU",  # Write Up
]
OS8_BLOCK_SIZE_BYTES = 384  # Block size (in bytes)


def os8_to_date(val: int) -> t.Optional[date]:
    """
    Translate OS-8 date to Python date
    """
    if val == 0:
        return None
    year = val & 0o7
    year += 1970
    day = (val >> 3) & 0o37
    month = (val >> 8) & 0o17
    if day == 0:
        day = 1
    if month == 0:
        month = 1
    try:
        return date(year, month, day)
    except:
        return None


def date_to_os8(d: t.Optional[date]) -> int:
    """
    Convert Python date to OS-8 date integer format.
    """
    if d is None:
        return 0
    year = (d.year - 1970) & 0o7
    val = (d.month << 8) | (d.day << 3) | year
    return val


def os8_canonical_filename(fullname: t.Optional[str], wildcard: bool = False) -> str:
    """
    Generate the canonical OS8 name
    """
    fullname = (fullname or "").upper()
    try:
        filename, extension = fullname.split(".", 1)
    except Exception:
        filename = fullname
        extension = "*" if wildcard else ""
    filename = (
        rad50_word12_to_asc(asc_to_rad50_word12(filename[0:2]))
        + rad50_word12_to_asc(asc_to_rad50_word12(filename[2:4]))
        + rad50_word12_to_asc(asc_to_rad50_word12(filename[4:6]))
    )
    extension = rad50_word12_to_asc(asc_to_rad50_word12(extension))
    return f"{filename}.{extension}"


def os8_split_fullname(
    partition: int, fullname: t.Optional[str], wildcard: bool = True
) -> t.Tuple[int, t.Optional[str]]:
    """
    Split the partition number from the fullname

    [1]filename.ext -> 1, filename.ext
    """
    if fullname:
        try:
            match = PARTITION_FULLNAME_RE.match(fullname)
            if match:
                partition_str, fullname = match.groups()
                partition = int(partition_str)
        except Exception:
            pass
        if fullname:
            fullname = os8_canonical_filename(fullname, wildcard=wildcard)
    return partition, fullname


def from_12bit_words_to_bytes(words: list[int], file_mode: str = ASCII) -> bytes:
    """
    Convert 12bit words to bytes

    http://www.bitsavers.org/pdf/dec/pdp8/os8/DEC-S8-OSSMB-A-D_OS8_v3ssup.pdf Pag 65
    """
    mask = 127 if file_mode == ASCII else 255
    data = bytearray()
    for i in range(0, len(words) - 1, 2):
        chr1 = words[i]
        try:
            chr2 = words[i + 1]
        except IndexError:
            chr2 = 0
        chr3 = ((chr2 >> 8) & 0o17) | ((chr1 >> 4) & 0o360)
        data.append(chr1 & mask)
        data.append(chr2 & mask)
        data.append(chr3 & mask)
    return bytes(data)


def from_bytes_to_12bit_words(byte_data: bytes, file_mode: str = "ASCII") -> t.List[int]:
    """
    Convert bytes to 12-bit words.
    """
    mask = 127 if file_mode == "ASCII" else 255
    words = []
    for i in range(0, len(byte_data), 3):
        chr1 = byte_data[i] & mask
        try:
            chr2 = byte_data[i + 1] & mask
        except IndexError:
            chr2 = 0
        try:
            chr3 = byte_data[i + 2] & mask
        except IndexError:
            chr3 = 0
        words.append(chr1 | ((chr3 & 0o360) << 4))
        words.append(chr2 | ((chr3 & 0o17) << 8))
    return words


def rad50_word12_to_asc(val: int) -> str:
    """
    Convert RAD50 12 bit word to 0-2 chars of ASCII
    """
    t = "".join(chr(c) if c > 0o40 else chr(c + 0o100) for c in [(val >> 6), (val & 0o77)])
    return t.replace("@", "")


def asc_to_rad50_word12(val: str) -> int:
    """
    Convert 0-2 chars of ASCII to RAD50 12 bit word
    """
    val = val.rjust(2, "@")
    c1 = ord(val[0])
    c2 = ord(val[1])
    if c1 < 0o100:
        c1 += 0o100
    if c2 < 0o100:
        c2 += 0o100
    return ((c1 & 0o77) << 6) | (c2 & 0o77)


def oct_dump(words: t.List[int], words_per_line: int = 8) -> None:
    """
    Display contents in octal
    """
    for i in range(0, len(words), words_per_line):
        line = words[i : i + words_per_line]
        ascii_str = "".join([chr(x) if 32 <= x <= 126 else "." for x in from_12bit_words_to_bytes(line)])
        oct_str = " ".join([f"{x:04o}" for x in line])
        sys.stdout.write(f"{i:08o}   {oct_str.ljust(5 * words_per_line)}  {ascii_str}\n")


class OS8File(AbstractFile):
    entry: "OS8DirectoryEntry"
    file_mode: str
    closed: bool
    size: int

    def __init__(self, entry: "OS8DirectoryEntry", file_mode: t.Optional[str] = None):
        self.entry = entry
        if file_mode is not None:
            self.file_mode = file_mode
        else:
            self.file_mode = ASCII if entry.extension.upper() in ASCII_EXTENSIONS else IMAGE
        self.closed = False
        self.size = entry.length * OS8_BLOCK_SIZE_BYTES

    def read_block(
        self,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> bytes:
        """
        Read block(s) of data from the file
        """
        if number_of_blocks == READ_FILE_FULL:
            number_of_blocks = self.entry.length
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        data = bytearray()
        for i in range(number_of_blocks):
            block_position = block_number + self.entry.file_position + i
            words = self.entry.segment.partition.read_words_block(block_position)
            t = from_12bit_words_to_bytes(words, self.file_mode)
            data.extend(t)
        return bytes(data)

    def write_block(
        self,
        buffer: bytes,
        block_number: int,
        number_of_blocks: int = 1,
    ) -> None:
        """
        Write block(s) of data to the file
        """
        if (
            self.closed
            or block_number < 0
            or number_of_blocks < 0
            or block_number + number_of_blocks > self.entry.length
        ):
            raise OSError(errno.EIO, os.strerror(errno.EIO))
        for i in range(number_of_blocks):
            block_position = block_number + self.entry.file_position + i
            data = buffer[i * OS8_BLOCK_SIZE_BYTES : (i + 1) * OS8_BLOCK_SIZE_BYTES]
            words = from_bytes_to_12bit_words(data, self.file_mode)
            self.entry.segment.partition.write_words_block(block_position, words)

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.entry.get_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return self.entry.get_block_size()

    def close(self) -> None:
        """
        Close the file
        """
        self.closed = True

    def __str__(self) -> str:
        return self.entry.fullname


class OS8DirectoryEntry(AbstractDirectoryEntry):
    """
    There are three types of file directory entries.
    - Permanent
    - Tentative
    - Empty

    A permanent entry appears as follows:

        +-----------------------------------+
      0 | File name char 1-2                |
      1 | File name char 3-4                |
      2 | File name char 5-6                |
      3 | Extension char 1-2                |
        +-----------------------------------+
      4 | Extra words                       |
        |.                                  |
        |.                                  |
        +-----------------------------------+
    N+4 | Minus file length in blocks       |
        +-----------------------------------+

    A tentative file entry appears as a permanent file entry with zero length.
    It is always immediately followed by an empty file entry.

    An empty entry appears as follows:

        +-----------------------------------+
      0 | Always 0                          |
      1 | Minus empty blocks                |
        +-----------------------------------+

    http://www.bitsavers.org/pdf/dec/pdp8/os8/DEC-S8-OSSMB-A-D_OS8_v3ssup.pdf Pag 63
    """

    segment: "OS8Segment"
    filename: str = ""
    extension: str = ""
    length: int = 0
    raw_creation_date: int = 0
    words: t.List[int] = []
    extra_words: t.List[int] = []
    file_position: int = 0
    empty_entry: bool = False

    def __init__(self, segment: "OS8Segment"):
        self.segment = segment

    @classmethod
    def read(cls, segment: "OS8Segment", words: t.List[int], position: int, file_position: int) -> "OS8DirectoryEntry":
        self = cls(segment)
        if words[0 + position] != 0:
            self.words = words[position : position + DIR_ENTRY_SIZE + segment.extra_words]
            n1 = words[0 + position]  # Filename char 1-2
            n2 = words[1 + position]  # Filename char 3-4
            n3 = words[2 + position]  # Filename char 5-6
            e1 = words[3 + position]  # Extension char 1-2
            self.empty_entry = False
            self.filename = rad50_word12_to_asc(n1) + rad50_word12_to_asc(n2) + rad50_word12_to_asc(n3)
            self.extension = rad50_word12_to_asc(e1)
            self.extra_words = words[4 + position : 4 + self.segment.extra_words + position]
            if self.segment.extra_words:
                # When extra words are used, the first one is used
                # to store the creation date
                self.raw_creation_date = self.extra_words[0]
            self.length = 999
            length = words[self.segment.extra_words + 4 + position]
            self.length = 0o10000 - length if length else length
        else:  # Empty entry
            length = 0o10000 - words[1 + position]
            self.words = words[position : position + EMPTY_DIR_ENTRY_SIZE]
            self.empty_entry = True
            self.filename = ""
            self.extension = ""
            self.extra_words = []
            self.length = length
        self.file_position = file_position
        return self

    def to_words(self) -> t.List[int]:
        """
        Write the directory entry
        """
        words = []
        if self.is_empty:
            words.append(0)
            words.append(0o10000 - self.length)
        else:
            words.append(asc_to_rad50_word12(self.filename[0:2]))
            words.append(asc_to_rad50_word12(self.filename[2:4]))
            words.append(asc_to_rad50_word12(self.filename[4:6]))
            words.append(asc_to_rad50_word12(self.extension))
            if self.segment.extra_words:
                self.extra_words[0] = self.raw_creation_date
            words.extend(self.extra_words)
            words.append(0o10000 - self.length)
        self.words = words
        return words

    @property
    def is_empty(self) -> bool:
        return self.empty_entry

    @property
    def is_tentative(self) -> bool:
        return (self.length == 0) and not self.is_empty

    @property
    def is_permanent(self) -> bool:
        return not self.is_empty and not self.is_tentative

    @property
    def directory_entry_len(self) -> int:
        """
        Length of this directory entry in words
        """
        if self.is_empty:
            return EMPTY_DIR_ENTRY_SIZE
        else:
            return DIR_ENTRY_SIZE + len(self.extra_words)

    @property
    def fullname(self) -> str:
        return f"{self.filename}.{self.extension}"

    @property
    def basename(self) -> str:
        return self.fullname

    def get_length(self) -> int:
        """
        Get the length in blocks
        """
        return self.length

    def get_size(self) -> int:
        """
        Get file size in bytes
        """
        return self.length * self.get_block_size()

    def get_block_size(self) -> int:
        """
        Get file block size in bytes
        """
        return OS8_BLOCK_SIZE_BYTES

    @property
    def creation_date(self) -> t.Optional[date]:
        return os8_to_date(self.raw_creation_date)

    def delete(self) -> bool:
        """
        Delete the file
        """
        self.empty_entry = True
        self.filename = ""
        self.extension = ""
        self.extra_words = []
        self.segment.compact()
        self.segment.write()
        self.words = self.to_words()
        return True

    def write(self) -> bool:
        """
        Write the directory entry
        """
        self.segment.write()
        return True

    def open(self, file_mode: t.Optional[str] = None) -> OS8File:
        """
        Open a file
        """
        return OS8File(self, file_mode)

    def __str__(self) -> str:
        attr = "TEMP" if self.is_tentative else "PERM" if self.is_permanent else "EMPTY"
        return f"{self.fullname:<11} {attr:<5} {self.creation_date or '          '} {self.length:>6} {self.file_position:6d}"

    def __repr__(self) -> str:
        return str(self)


class OS8Segment:
    """
    Volume Directory Segment

    OS/8 - Pag 62
    http://www.bitsavers.org/pdf/dec/pdp8/os8/DEC-S8-OSSMB-A-D_OS8_v3ssup.pdf

    COS-310 - Pag 5
    https://svn.so-much-stuff.com/svn/trunk/pdp8/Websites/pdp-8.org/scans/highgate/cos310-docs/aa-d647a-ta-app.pdf

        +-----------------------------------+
      0 | Minus number of entries           |   5 words header
      1 | First file starting block         |
      2 | Link to next segment              |
      3 | Point to last tentative file word |  (unused in COS)
      4 | Minus number of extra words       |  (always -1 for COS)
        +-----------------------------------+
      5 |Entries                            |
        |.                                  |
    255 |.                                  |
        +-----------------------------------+
    """

    partition: "OS8Partition"
    # Block number of this directory segment
    block_number = 0
    # Block number where the stored data identified by this segment begins
    data_block_number = 0
    # Next directory segment block number
    next_block_number = 0
    # Points to last word of tentative file entry in this segment
    tentative_last_word = 0
    # Number of extra words per directory entry
    extra_words = 0
    # Directory entries
    entries_list: t.List["OS8DirectoryEntry"] = []

    def __init__(self, partition: "OS8Partition"):
        self.partition = partition
        self.entries_list = []

    @classmethod
    def read(cls, partition: "OS8Partition", block_number: int) -> "OS8Segment":
        """
        Read a Volume Directory Segment from disk
        """
        self = cls(partition)
        data = self.partition.read_words_block(block_number)
        self.block_number = block_number
        number_of_entries = 0o10000 - data[0]  # Minus the number of entries
        self.data_block_number = data[1]  # Block number where the stored data begins
        self.next_block_number = data[2]  # Link to next segment
        self.tentative_last_word = data[3]  # Link to next segment
        self.extra_words = 0o10000 - data[4]  # Minus the number of extra words
        self.entries_list = []

        file_position = self.data_block_number
        position = DIRECTORY_SEGMENT_HEADER_SIZE
        for _ in range(0, number_of_entries):
            dir_entry = self.partition.fs.directory_entry_class.read(self, data, position, file_position)
            file_position = file_position + dir_entry.length
            position = position + dir_entry.directory_entry_len
            self.entries_list.append(dir_entry)
        return self

    def compact(self) -> None:
        """
        Compact multiple unused entries
        """
        prev_empty_entry = None
        new_entries_list = []
        for entry in self.entries_list:
            if not entry.is_empty:
                prev_empty_entry = None
                new_entries_list.append(entry)
            elif prev_empty_entry is None:
                prev_empty_entry = entry
                new_entries_list.append(entry)
            else:
                prev_empty_entry.length = prev_empty_entry.length + entry.length
        self.entries_list = new_entries_list

    def write(self) -> None:
        """
        Write the Volume Directory Segment
        """
        words = []
        words.append(0o10000 - self.number_of_entries)
        words.append(self.data_block_number)
        words.append(self.next_block_number)
        words.append(self.tentative_last_word)
        words.append(0o10000 - self.extra_words)
        for entry in self.entries_list:
            words.extend(entry.to_words())
        words += [0] * (DIRECTORY_SEGMENT_SIZE - len(words))
        self.partition.write_words_block(self.block_number, words)

    @property
    def number_of_entries(self) -> int:
        """
        Number of directory entries in this segment
        """
        return len(self.entries_list)

    @property
    def max_entries(self) -> int:
        """
        Max directory entries
        """
        return (DIRECTORY_SEGMENT_SIZE - DIRECTORY_SEGMENT_HEADER_SIZE) // (DIR_ENTRY_SIZE + self.extra_words) - 1

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["OS8DirectoryEntry"]:
        for entry in self.entries_list:
            if filename_match(entry.basename, pattern, wildcard):
                if not include_all and (entry.is_empty or entry.is_tentative):
                    continue
                yield entry

    def insert_empty_entry_after(self, entry: "OS8DirectoryEntry", entry_number: int, length: int) -> None:
        """
        Insert an empty entry after the given entry
        If the entry length is equal to the requested length, don't create the new empty entity
        """
        if entry.length == length:
            return
        # Create new empty space entry
        new_entry = self.partition.fs.directory_entry_class(self)
        new_entry.empty_entry = True
        new_entry.length = entry.length - length
        new_entry.file_position = entry.file_position + length
        entry.length = length
        self.entries_list.insert(entry_number + 1, new_entry)
        entry.segment.write()

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        free_blocks = 0
        for entry in self.entries_list:
            if entry.is_empty:
                free_blocks += entry.length
        return free_blocks

    def __str__(self) -> str:
        buf = io.StringIO()
        buf.write("\n*Segment\n")
        buf.write(f"Block number:          {self.block_number:>5}\n")
        buf.write(f"Number of entries:     {self.number_of_entries:>5}\n")
        buf.write(f"Data block:            {self.data_block_number:>5}\n")
        buf.write(f"Next dir segment:      {self.next_block_number:>5}\n")
        buf.write(f"Tentative last word:   {self.tentative_last_word:>5}\n")
        buf.write(f"Extra words:           {self.extra_words:>5}\n")
        buf.write(f"Max entries:           {self.max_entries:>5}\n\n")
        buf.write("Num  Filename    Type  Date       Length  Block\n")
        buf.write("---  --------    ----  ----       ------  -----\n")
        for i, x in enumerate(self.entries_list):
            buf.write(f"{i:02d}#  {x}\n")
        return buf.getvalue()


class OS8Partition:
    """
    OS/8 partition

        +-----------------------------------+
      0 | System bootstrap                  |
        +-----------------------------------+
      1 | Directory segments                |
        |.                                  |
      7 |.                                  |
        +-----------------------------------+
        | OS/8 System (Optional)            |
        |.                                  |
        +-----------------------------------+
        | File storage                      |
        |.                                  |
        |.                                  |
        +-----------------------------------+
    """

    fs: "OS8Filesystem"
    # partition number
    partition_number: int
    # partition size
    partition_size: int
    # Block number of the first block of this partition
    base_block_number: int

    def __init__(self, fs: "OS8Filesystem", partition_number: int):
        self.fs = fs
        self.partition_number = partition_number
        self.partition_size = fs.partition_size
        self.base_block_number = partition_number * fs.partition_size

    def read_words_block(self, block_number: int) -> t.List[int]:
        """
        Read a 512 bytes block as 256 12bit words
        """
        return self.fs.read_words_block(block_number + self.base_block_number)

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write 256 12bit words as 512 bytes block
        """
        return self.fs.write_words_block(block_number + self.base_block_number, words)

    def read_dir_segments(self) -> t.Iterator["OS8Segment"]:
        """
        Read all directory segments of this partition
        """
        next_block_number = DIRECTORY_SEGMENT_START
        while next_block_number != 0:
            segment = OS8Segment.read(self, next_block_number)
            next_block_number = segment.next_block_number
            yield segment

    def get_file_entry(self, fullname: str) -> OS8DirectoryEntry:  # fullname=filename+ext
        """
        Get the directory entry for a file in this partition
        """
        for segment in self.read_dir_segments():
            for entry in segment.entries_list:
                if entry.fullname == fullname and entry.is_permanent:
                    return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def split_segment(self, entry: OS8DirectoryEntry) -> bool:
        # entry is the last entry of the old_segment, new new segment will contain all the entries after that
        old_segment = entry.segment
        # find the new segment number
        segments = list(self.read_dir_segments())
        sn = [x.block_number for x in segments]
        block_number = None
        for i in range(0, NUM_OF_SEGMENTS):
            block_number_t = self.base_block_number + DIRECTORY_SEGMENT_START + i
            if block_number_t not in sn:
                block_number = block_number_t
                break
        if block_number is None:
            return False
        # create the new segment
        segment = OS8Segment(self)
        segment.block_number = block_number
        segment.data_block_number = entry.file_position + entry.length
        segment.next_block_number = old_segment.next_block_number
        segment.tentative_last_word = 0
        segment.extra_words = segments[0].extra_words
        # set the next segment of the last segment
        old_segment.next_block_number = segment.block_number

        entry_position = -1
        for i, e in enumerate(old_segment.entries_list):
            if entry == e:
                entry_position = i
        if entry_position == 1:
            return False
        segment.entries_list = old_segment.entries_list[entry_position + 1 :]
        old_segment.entries_list = old_segment.entries_list[: entry_position + 1]
        old_segment.write()
        segment.data_block_number = entry.file_position + entry.length
        segment.write()
        return True

    def search_empty_entry(self, length: int) -> t.Tuple[t.Optional[OS8DirectoryEntry], int]:
        # Search for an empty entry to be split
        entry: t.Optional[OS8DirectoryEntry] = None
        entry_number: int = -1
        for segment in self.read_dir_segments():
            for i, e in enumerate(segment.entries_list):
                if e.is_empty and e.length >= length:
                    if entry is None or entry.length > e.length:
                        entry = e
                        entry_number = i
                        if entry.length == length:
                            break
        return entry, entry_number

    def allocate_space(
        self,
        fullname: str,  # fullname=filename+ext, length in blocks
        length: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
    ) -> OS8DirectoryEntry:
        """
        Allocate space for a new file
        """
        # Search for an empty entry to be split
        entry, entry_number = self.search_empty_entry(length)
        if entry is None:
            raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC), fullname)
        # If the entry length is equal to the requested length, don't create the new empty entity
        if entry.length != length:
            if len(entry.segment.entries_list) >= entry.segment.max_entries:
                if not self.split_segment(entry):
                    raise OSError(errno.ENOSPC, os.strerror(errno.ENOSPC), fullname)
            entry.segment.insert_empty_entry_after(entry, entry_number, length)
        # Fill the entry
        split = os.path.splitext(fullname.upper())
        entry.empty_entry = False
        entry.extra_words = [0] * entry.segment.extra_words
        entry.filename = split[0]
        entry.extension = split[1] and split[1][1:] or ""
        entry.raw_creation_date = date_to_os8(creation_date)
        entry.length = length
        # Write the segment
        entry.segment.write()
        return entry

    def create_file(
        self,
        fullname: str,
        number_of_blocks: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # t.optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[OS8DirectoryEntry]:
        try:
            self.get_file_entry(fullname).delete()
        except FileNotFoundError:
            pass
        return self.allocate_space(fullname, number_of_blocks, creation_date)

    def initialize(self, **kwargs: t.Union[bool, str]) -> None:
        """
        Create an empty OS/8 partition
        """
        # Directory segment
        segment = OS8Segment(self)
        segment.block_number = DIRECTORY_SEGMENT_START
        segment.data_block_number = segment.block_number + NUM_OF_SEGMENTS
        segment.next_block_number = 0
        segment.tentative_last_word = 0
        segment.extra_words = 1
        # Empty directory entry
        length = self.partition_size - segment.data_block_number
        if getattr(self.fs.dev, "is_rx", False):
            length = length - RX_SECTOR_TRACK * self.fs.dev.sector_size // BLOCK_SIZE
        dir_entry = self.fs.directory_entry_class(segment)
        dir_entry.empty_entry = True
        dir_entry.filename = ""
        dir_entry.extension = ""
        dir_entry.length = length
        segment.entries_list.append(dir_entry)
        segment.write()

    def free(self) -> int:
        """
        Count the number of free blocks
        """
        free_blocks = 0
        for segment in self.read_dir_segments():
            free_blocks += segment.free()
        return free_blocks

    def __str__(self) -> str:
        buf = io.StringIO()
        buf.write("\n*Partition\n")
        buf.write(f"Partition number:         {self.partition_number:>5}\n")
        buf.write(f"Partition size:           {self.partition_size:>5}\n")
        buf.write(f"Partition starting block: {self.base_block_number:>5}\n")
        return buf.getvalue()


class OS8Filesystem(AbstractFilesystem):
    """
    OS/8 Filesystem

    OS/8 FILE STRUCTURES
    http://www.bitsavers.org/pdf/dec/pdp8/os8/DEC-S8-OSSMB-A-D_OS8_v3ssup.pdf Pag 62
    """

    fs_name = "os8"
    fs_description = "PDP-8 OS/8"
    fs_platforms = ["pdp-8"]
    dev: BlockDevice12Bit

    current_partition: int  # Current partition
    number_of_blocks: int  # Number of blocks
    directory_entry_class: t.Type[OS8DirectoryEntry] = OS8DirectoryEntry

    def __init__(self, file_or_device: t.Union["AbstractFile", "AbstractDevice"], device_type: t.Union[bool, str] = ""):
        if isinstance(file_or_device, AbstractFile):
            if device_type == DECTAPE:
                self.dev = DECtape12Bit(file_or_device)
            else:
                self.dev = RXBlockDevice12Bit(file_or_device)
        elif isinstance(file_or_device, BlockDevice12Bit):
            self.dev = file_or_device
        else:
            raise OSError(errno.EIO, f"Invalid device type for {self.fs_description} filesystem")

    @property
    def num_of_partitions(self) -> int:
        """Get the number of partitions"""
        return 1 + (self.number_of_blocks - 1) // 0o10000

    @property
    def partition_size(self) -> int:  # Size of each partition
        """Get the size of each partition"""
        return self.number_of_blocks // self.num_of_partitions

    def get_partition(self, partition_number: int) -> OS8Partition:
        """
        Get a partition by number
        """
        try:
            if partition_number is None:
                partition_number = self.current_partition
            if partition_number < 0 or partition_number >= self.num_of_partitions:
                raise ValueError
            return OS8Partition(self, partition_number)
        except:
            raise FileNotFoundError(errno.ENOENT, "Partition not found", partition_number)

    @classmethod
    def mount(
        cls,
        file_or_dev: t.Union["AbstractFile", "AbstractDevice"],
        strict: t.Union[bool, str] = True,
        device_type: t.Union[bool, str] = "",
        **kwargs: t.Union[bool, str],
    ) -> "OS8Filesystem":
        """
        Mount the OS/8 filesystem from a file
        """
        self = cls(file_or_dev, device_type)
        self.current_partition = 0
        self.number_of_blocks = self.dev.get_size() // BLOCK_SIZE
        if strict:
            try:
                # Check if the filesystem is valid by reading the directory entries
                self.check_os8()
            except Exception:
                if isinstance(file_or_dev, AbstractFile) and device_type == DECTAPE:
                    # Try to read the device not as DECtape
                    self.dev = RXBlockDevice12Bit(file_or_dev)
                    self.check_os8()
                else:
                    raise
        return self

    def check_os8(self) -> None:
        """
        Check if the filesystem is a valid OS/8 filesystem
        """
        # Check if the filesystem is valid by reading the directory entries
        for _ in self.entries_list:
            pass

    def read_words_block(self, block_number: int) -> t.List[int]:
        """
        Read a block as 256 12bit words
        """
        return self.dev.read_words_block(block_number)

    def write_words_block(
        self,
        block_number: int,
        words: t.List[int],
    ) -> None:
        """
        Write 256 12bit words as a block
        """
        self.dev.write_words_block(block_number, words)

    def filter_entries_list(
        self,
        pattern: t.Optional[str],
        include_all: bool = False,
        expand: bool = True,
        wildcard: bool = True,
    ) -> t.Iterator["OS8DirectoryEntry"]:
        partition = self.current_partition
        if pattern:
            partition, pattern = os8_split_fullname(partition, pattern, wildcard)
        for segment in self.get_partition(partition).read_dir_segments():
            yield from segment.filter_entries_list(
                pattern,
                include_all=include_all,
                expand=expand,
                wildcard=wildcard,
            )

    @property
    def entries_list(self) -> t.Iterator["OS8DirectoryEntry"]:
        for segment in self.get_partition(self.current_partition).read_dir_segments():
            for entry in segment.entries_list:
                yield entry

    def get_file_entry(self, fullname: str) -> OS8DirectoryEntry:  # fullname=filename+ext
        partition = self.current_partition
        if fullname:
            partition, fullname = os8_split_fullname(partition, fullname)  # type: ignore
        for segment in self.get_partition(partition).read_dir_segments():
            for entry in segment.entries_list:
                if entry.fullname == fullname and entry.is_permanent:
                    return entry
        raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), fullname)

    def write_bytes(
        self,
        fullname: str,
        content: bytes,
        creation_date: t.Optional[date] = None,
        file_type: t.Optional[str] = None,
        file_mode: t.Optional[str] = None,
    ) -> None:
        number_of_blocks = int(math.ceil(len(content) / OS8_BLOCK_SIZE_BYTES))
        entry = self.create_file(fullname, number_of_blocks, creation_date, file_type)
        if entry is not None:
            content = content + (b"\0" * OS8_BLOCK_SIZE_BYTES)
            f = entry.open(file_mode)
            try:
                f.write_block(content, block_number=0, number_of_blocks=entry.length)
            finally:
                f.close()

    def create_file(
        self,
        fullname: str,
        length: int,  # length in blocks
        creation_date: t.Optional[date] = None,  # optional creation date
        file_type: t.Optional[str] = None,
    ) -> t.Optional[OS8DirectoryEntry]:
        partition, fullname = os8_split_fullname(self.current_partition, fullname, wildcard=False)  # type: ignore
        return self.get_partition(partition).create_file(fullname, length, creation_date, file_type)

    def chdir(self, fullname: str) -> bool:
        try:
            partition = int(fullname)
        except:
            return False
        if partition < 0 or partition >= self.num_of_partitions:
            return False
        self.current_partition = int(fullname)
        return True

    def isdir(self, fullname: str) -> bool:
        return False

    def dir(self, volume_id: str, pattern: t.Optional[str], options: t.Dict[str, bool]) -> None:
        i = 0
        files = 0
        blocks = 0
        partition = self.current_partition
        if pattern:
            partition, pattern = os8_split_fullname(partition, pattern, wildcard=True)
        part = self.get_partition(partition)
        for segment in part.read_dir_segments():
            for x in segment.filter_entries_list(pattern, include_all=True):
                if x.is_empty or x.is_tentative:
                    if not options.get("full"):
                        continue
                    i = i + 1
                    fullname = "<EMPTY>  "
                    file_date = ""
                else:
                    i = i + 1
                    fullname = f"{x.filename:<6}.{x.extension:<2}"
                    if options.get("brief"):
                        # Lists only file names and file types
                        sys.stdout.write(f"{fullname}\n")
                    else:
                        files = files + 1
                        blocks = blocks + x.length
                        file_date = x.creation_date and x.creation_date.strftime("%d-%b-%y").upper() or ""
                sys.stdout.write(f"{fullname} {x.file_position:04o} {x.length:>3} {file_date:<9}")
                if i % 3 == 0:
                    sys.stdout.write("\n")
                else:
                    sys.stdout.write("  ")
        if not options.get("brief"):
            if i % 3 != 0:
                sys.stdout.write("\n")
            unused = part.free()
            sys.stdout.write(f"\n{files:>4} FILES IN {blocks:>4} BLOCKS - {unused:>4} FREE BLOCKS\n")

    def examine(self, arg: t.Optional[str], options: t.Dict[str, t.Union[bool, str]]) -> None:
        if arg:
            sys.stdout.write("Filename    Type  Date       Length  Block\n")
            sys.stdout.write("--------    ----  ----       ------  -----\n")
            for entry in self.filter_entries_list(arg, include_all=True):
                if options.get("full"):
                    words = [f"{x:04o}" for x in entry.words]
                    sys.stdout.write(f"{entry}  {words}\n")
                else:
                    sys.stdout.write(f"{entry}\n")
        else:
            is_rx = getattr(self.dev, "is_rx", False)
            sys.stdout.write(f"Number of partitions:     {self.num_of_partitions}\n")
            sys.stdout.write(f"Size of each partition:   {self.partition_size}\n")
            sys.stdout.write(f"Is RX01/RX02:             {'YES' if is_rx else 'NO'}\n")
            for partition_number in range(0, self.num_of_partitions):
                partition = self.get_partition(partition_number)
                sys.stdout.write(f"{partition}\n")
                for segment in partition.read_dir_segments():
                    sys.stdout.write(f"{segment}\n")

    def dump(self, fullname: t.Optional[str], start: t.Optional[int] = None, end: t.Optional[int] = None) -> None:
        """Dump the content of a file or a range of blocks"""
        if fullname:
            entry = self.get_file_entry(fullname)
            if start is None:
                start = 0
            if end is None or end > entry.get_length() - 1:
                end = entry.get_length() - 1
            for block_number in range(start, end + 1):
                words = self.read_words_block(entry.file_position + block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)
        else:
            if start is None:
                start = 0
                if end is None:  # full disk
                    end = self.get_size() // BLOCK_SIZE - 1
            elif end is None:  # one single block
                end = start
            for block_number in range(start, end + 1):
                words = self.read_words_block(block_number)
                sys.stdout.write(f"\nBLOCK NUMBER   {block_number:08}\n")
                oct_dump(words)

    @classmethod
    def initialize(
        cls, file_or_dev: t.Union["AbstractFile", "AbstractDevice"], **kwargs: t.Union[bool, str]
    ) -> "OS8Filesystem":
        """
        Create an empty OS/8 filesystem
        """
        self = cls(file_or_dev)
        self.current_partition = 0
        assert isinstance(self.dev, BlockDevice12Bit), "OS/8 filesystem requires a 12-bit block device"
        self.number_of_blocks = self.dev.get_size() // BLOCK_SIZE
        # Initialize the partitions
        for partition_number in range(0, self.num_of_partitions):
            partition = OS8Partition(self, partition_number)
            partition.initialize(**kwargs)
        return self

    def get_size(self) -> int:
        """
        Get filesystem size in bytes
        """
        return self.dev.get_size()

    def close(self) -> None:
        self.dev.close()

    def get_pwd(self) -> str:
        if self.current_partition == 0:
            return ""
        else:
            return f"[{self.current_partition}]"
